from email.mime import application
import time

from ..request import Request
from ..response import Response
from .Provider import Provider

from ..presets.PresetsCapsule import PresetsCapsule
from ..presets import Tailwind, Vue, React, Bootstrap


class FrameworkProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        # @M5 remove this and add PresetsProvider in default project
        presets = PresetsCapsule()
        presets.add(Bootstrap())
        presets.add(Tailwind())
        presets.add(Vue())
        presets.add(React())
        self.application.bind("presets", presets)

    def boot(self):
        request = Request(self.application.make("environ"))
        request.app = self.application
        if self.application.has("activate.subdomains") and self.application.make(
            "activate.subdomains"
        ):
            request.activate_subdomains()
        self.application.bind("request", request)
        self.application.bind("response", Response(self.application))

        self.application.bind("start_time", time.time())



from ...request import Request
from ...controllers import Controller
from ...response import Response


class ExceptioniteController(Controller):
    def run_action(self, request: Request, response: Response):
        handler = request.app.make("exception_handler").get_driver("exceptionite")
        data = handler.run_action(request.input("action_id"), request.input("options"))
        try:
            return response.json({"message": "ok", "data": data}, 200)
        except:  # noqa: E722
            return response.json({"message": "An error happened", "data": data}, 400)



class ExceptionHandler:
    def __init__(self, application, driver_config=None):
        self.application = application
        self.drivers = {}
        self.driver_config = driver_config or {}
        self.options = {}

    def set_options(self, options):
        self.options = options
        return self

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})

    def set_configuration(self, config):
        self.driver_config = config
        return self

    def get_driver(self, name=None):
        if name is None:
            return self.drivers[self.driver_config.get("default")]
        return self.drivers[name]

    def get_config_options(self, driver=None):
        if driver is None:
            return self.driver_config[self.driver_config.get("default")]

        return self.driver_config.get(driver, {})

    def handle(self, exception):
        response = self.application.make("response")
        request = self.application.make("request")

        self.application.make("event").fire(
            f"masonite.exception.{exception.__class__.__name__}", exception
        )

        # if an exception handler is registered for this exception, use it instead
        if self.application.has(f"{exception.__class__.__name__}Handler"):
            return self.application.make(
                f"{exception.__class__.__name__}Handler"
            ).handle(exception)

        # handle exception in production
        if not self.application.is_debug():
            # for HTTP error codes (500, 404, 403...) a specific page should be displayed
            # if a renderable exception is raised let it be displayed
            if hasattr(exception, "is_http_exception") or hasattr(
                exception, "get_response"
            ):
                return self.application.make("HttpExceptionHandler").handle(exception)

            # else fallback to an unknown exception that should be displayed as a 500 error
            exception.get_status = lambda: 500
            exception.get_response = lambda: str(exception) or "Unknown error"
            return self.application.make("HttpExceptionHandler").handle(exception)

        # handle exception in development mode with Exceptionite
        exceptionite = self.get_driver("exceptionite")
        exceptionite.start(exception)
        exceptionite.render("terminal")

        if request.accepts_json():
            return response.view(exceptionite.render("json"), status=500)
        else:
            return response.view(exceptionite.render("web"), status=500)



from tests import TestCase
from unittest.mock import patch
from src.masonite.notification import Notification, Notifiable, Sms, Textable
from src.masonite.exceptions import NotificationException

from masoniteorm.models import Model


class User(Model, Notifiable):
    """User Model"""

    __fillable__ = ["name", "email", "password", "phone"]

    def route_notification_for_vonage(self):
        return "+33123456789"


class WelcomeUserNotification(Notification, Textable):
    def to_vonage(self, notifiable):
        return self.text_message("Welcome !").from_("123456")

    def via(self, notifiable):
        return ["vonage"]


class WelcomeNotification(Notification):
    def to_vonage(self, notifiable):
        return Sms().text("Welcome !").from_("123456")

    def via(self, notifiable):
        return ["vonage"]

    def should_send(self):
        return True


class OtherNotification(Notification):
    def to_vonage(self, notifiable):
        return Sms().text("Welcome !")

    def via(self, notifiable):
        return ["vonage"]


class VonageAPIMock(object):
    @staticmethod
    def send_success():
        return {"hoho": "hihi", "message-count": 1, "messages": [{"status": "0"}]}

    @staticmethod
    def send_error(error="Missing api_key", status=2):
        return {
            "message-count": 1,
            "messages": [{"status": str(status), "error-text": error}],
        }


class TestVonageDriver(TestCase):
    def setUp(self):
        super().setUp()
        self.notification = self.application.make("notification")

    def test_sending_without_credentials(self):
        with self.assertRaises(NotificationException) as e:
            self.notification.route("vonage", "+33123456789").send(
                WelcomeNotification()
            )
        error_message = str(e.exception)
        self.assertIn("Code [2]", error_message)

    def test_send_to_anonymous(self):
        with patch("vonage.sms.Sms") as MockSmsClass:
            MockSmsClass.return_value.send_message.return_value = (
                VonageAPIMock().send_success()
            )
            self.notification.route("vonage", "+33123456789").send(
                WelcomeNotification()
            )

    def test_send_to_notifiable(self):
        with patch("vonage.sms.Sms") as MockSmsClass:
            MockSmsClass.return_value.send_message.return_value = (
                VonageAPIMock().send_success()
            )
            user = User.find(1)
            user.notify(WelcomeUserNotification())

    def test_send_to_notifiable_with_route_notification_for(self):
        with patch("vonage.sms.Sms") as MockSmsClass:
            MockSmsClass.return_value.send_message.return_value = (
                VonageAPIMock().send_success()
            )
            user = User.find(1)
            user.notify(WelcomeNotification())

    def test_global_send_from_is_used_when_not_specified(self):
        notifiable = self.notification.route("vonage", "+33123456789")
        sms = self.notification.get_driver("vonage").build(
            notifiable, OtherNotification()
        )
        self.assertEqual(sms._from, "+33000000000")



import unittest
from src.masonite.headers import HeaderBag, Header


class TestHeaders(unittest.TestCase):
    def test_headers_can_be_registered(self):
        bag = HeaderBag()
        bag.add(Header("Content-Type", "application/json"))
        self.assertTrue(bag["CONTENT_TYPE"])

    def test_headers_can_be_rendered(self):
        bag = HeaderBag()
        bag.add(Header("content-type", "application/json"))
        bag.add(Header("x-forwarded-for", "127.0.0.1"))
        self.assertEqual(
            bag.render(),
            [("Content-Type", "application/json"), ("X-Forwarded-For", "127.0.0.1")],
        )

    def test_headers_can_check_with_in(self):
        bag = HeaderBag()
        bag.add(Header("Content-Type", "application/json"))

        self.assertTrue("Content-Type" in bag)

    def test_headers_can_be_retrieved(self):
        bag = HeaderBag()
        bag.add(Header("Content-Type", "application/json"))
        self.assertEqual(bag.get("content-type").value, "application/json")
        self.assertEqual(bag.get("Content-Type").value, "application/json")



from src.masonite.middleware.route.IpMiddleware import IpMiddleware
from tests import TestCase
from src.masonite.request import Request
from src.masonite.utils.http import generate_wsgi


class TestRequest(TestCase):
    def test_request_can_get_path(self):
        request = Request(generate_wsgi(path="/test"))
        self.assertEqual(request.get_path(), "/test")
        self.assertEqual(request.get_request_method(), "GET")

    def test_request_contains(self):
        request = Request(generate_wsgi(path="/test"))
        self.assertTrue(request.contains("/test"))

        request = Request(generate_wsgi(path="/test/user"))
        self.assertTrue(request.contains("/test/*"))

        request = Request(generate_wsgi(path="/test/admin/user"))
        self.assertTrue(request.contains("/test/*/user"))

        request = Request(generate_wsgi(path="/test/admin/user"))
        self.assertTrue(request.contains("*"))

    def test_request_ip(self):
        request = self.make_request(
            {
                # private or reserved
                "HTTP_CLIENT_IP": "172.16.0.2,255.255.255.255",
                # should be okay
                "HTTP_X_FORWARDED": "93.16.100.10",
                "REMOTE_ADDR": "192.167.2.1",
            }
        )
        response = self.make_response()
        request = IpMiddleware().before(request, response)
        self.assertEqual(request.ip(), "93.16.100.10")

        request = self.make_request(
            {
                "REMOTE_ADDR": "127.0.0.1",
            }
        )
        request = IpMiddleware().before(request, response)
        self.assertEqual(request.ip(), "127.0.0.1")

    def test_requests_accepts_json(self):
        request = self.make_request({"HTTP_ACCEPT": "application/json"})
        self.assertTrue(request.accepts_json())
        request = self.make_request({"HTTP_ACCEPT": "text/html"})
        self.assertFalse(request.accepts_json())

    def test_getting_request_method(self):
        request = self.make_request(method="POST")
        self.assertEqual(request.get_request_method(), "POST")

    def test_overriding_request_method_in_form(self):
        request = self.make_request(post_data={"__method": "PATCH"}, method="POST")
        self.assertEqual(request.get_request_method(), "PATCH")



from typing import TYPE_CHECKING, Callable, List, Tuple, Any

if TYPE_CHECKING:
    from ..authorization import AuthorizationResponse, Policy, Gate as GateObject

class Gate:
    """Gate facade."""

    def define(permission: str, condition: Callable): ...
    def register_policies(policies: List[Tuple[Any, Policy]]) -> "Gate": ...
    def get_policy_for(instance_or_class: "str|Any") -> "None|Policy": ...
    def before(before_callback: Callable): ...
    def after(after_callback: Callable): ...
    def allows(permission: str, *args) -> bool: ...
    def denies(permission, *args) -> bool: ...
    def has(permission: str) -> bool: ...
    def for_user(user: Any) -> GateObject: ...
    def any(permissions: List[str], *args) -> bool:
        """Check that every of those permissions are allowed."""
        ...
    def none(permissions: List[str], *args) -> bool:
        """Check that none of those permissions are allowed."""
        ...
    def authorize(permission: str, *args) -> bool: ...
    def inspect(permission: str, *args) -> "bool|AuthorizationResponse":
        """Get permission checks results for the given user then builds and returns an
        authorization response."""
        ...
    def check(permission: str, *args):
        """The core of the authorization class. Run before() checks, permission check and then
        after() checks."""
        ...



"""Mail notification driver."""

from .BaseDriver import BaseDriver


class MailDriver(BaseDriver):
    def __init__(self, application):
        self.application = application
        self.options = {}

    def set_options(self, options):
        self.options = options
        return self

    def send(self, notifiable, notification):
        """Used to send the email."""
        mailable = self.get_data("mail", notifiable, notification)
        if not mailable._to:
            recipients = notifiable.route_notification_for("mail")
            mailable = mailable.to(recipients)
        return self.application.make("mail").mailable(mailable).send()



import inspect
import warnings

from .exceptions import DumpException


warnings.warn(
    "DD class will be removed in Masonite 5. Please use Dump facade instead.",
    DeprecationWarning,
)


class DD:
    def __init__(self, container):
        self.app = container

    def die_and_dump(self, *args):
        """Dump all provided args and die, ie raise a DumpException."""
        self.dump(*args)
        raise DumpException

    def dump(self, *args):
        """Dump all provided args and let flow continue. This does not raise a DumpException."""
        print(
            inspect.stack()[1].function,
            inspect.stack()[1].filename,
            inspect.stack()[1].lineno,
        )
        if self.app.has("ObjDumpList"):
            dump_list = self.app.make("ObjDumpList")
        else:
            dump_list = []
        start = len(dump_list)
        for i, obj in enumerate(args):
            dump_name = f"ObjDump{start + i}"
            self.app.bind(dump_name, obj)
            dump_list.append(dump_name)
        self.app.bind("ObjDumpList", dump_list)



from typing import Any

class Config:
    """Configuration facades to manage Masonite config files."""

    def get(key: str, default: None) -> Any:
        """Get given key in config, can use a dotted path."""
        ...
    def set(key: str, value: Any) -> None:
        """Override config for given key."""
        ...
    def all(self) -> dict:
        """Get all config object."""
        ...



import pendulum
from tests import TestCase

from src.masonite.utils.time import (
    migration_timestamp,
    parse_human_time,
    cookie_expire_time,
)


class TestTimeUtils(TestCase):
    def tearDown(self):
        super().tearDown()
        self.restoreTime()

    def test_parse_human_time_now(self):
        ref_time = pendulum.datetime(2021, 1, 1)
        self.fakeTime(ref_time)
        instance = parse_human_time("now")
        self.assertEqual(ref_time, instance)

    def test_parse_human_time_expired(self):
        self.fakeTime(pendulum.datetime(2021, 1, 1))
        instance = parse_human_time("expired")
        self.assertEqual(pendulum.datetime(2001, 1, 1), instance)

    def test_parse_human_time(self):
        self.fakeTime(pendulum.datetime(2021, 1, 1, 12, 0, 0))
        self.assertEqual(
            pendulum.datetime(2021, 1, 1, 12, 0, 2), parse_human_time("2 seconds")
        )
        self.assertEqual(
            pendulum.datetime(2021, 1, 1, 12, 2, 0), parse_human_time("2 minutes")
        )
        self.assertEqual(
            pendulum.datetime(2021, 1, 1, 14, 0, 0), parse_human_time("2 hour")
        )
        self.assertEqual(
            pendulum.datetime(2021, 1, 2, 12, 0, 0), parse_human_time("1 day")
        )
        self.assertEqual(
            pendulum.datetime(2021, 1, 15, 12, 0, 0), parse_human_time("2 weeks")
        )
        self.assertEqual(
            pendulum.datetime(2021, 4, 1, 12, 0, 0), parse_human_time("3 months")
        )
        self.assertEqual(
            pendulum.datetime(2030, 1, 1, 12, 0, 0), parse_human_time("9 years")
        )

        self.assertEqual(None, parse_human_time("10 nanoseconds"))

    def test_cookie_expire_time(self):
        self.fakeTime(pendulum.datetime(2021, 1, 21, 7, 28, 0))
        expiration_time_str = cookie_expire_time("7 days")
        self.assertEqual(expiration_time_str, "Thu, 28 Jan 2021 07:28:00")

    def test_migration_timestamp(self):
        self.fakeTime(pendulum.datetime(2021, 10, 25, 8, 12, 54))
        self.assertEqual(migration_timestamp(), "2021_10_25_081254")



"""Cache Config"""
import os
from src.masonite.utils.location import base_path

DISKS = {
    "default": "local",
    "local": {
        "driver": "file",
        "path": base_path("storage/framework/filesystem"),
    },
    "s3": {
        "driver": "s3",
        "client": os.getenv("AWS_CLIENT"),
        "secret": os.getenv("AWS_SECRET"),
        "bucket": os.getenv("AWS_BUCKET"),
    },
}



from typing import Any

class Response:
    """Response facade."""

    def json(payload: Any, status: int = 200) -> bytes:
        """Set the response as a JSON response."""
        ...
    def make_headers(content_type: str = "text/html; charset=utf-8") -> None:
        """Recompute Content-Length of the response after modyifing it."""
        ...
    def header(name: str, value: str = None) -> "None|str":
        """Get response header for the given name if no value provided or add headers to response."""
        ...
    def get_headers(self) -> list:
        """Get all response headers."""
        ...
    def cookie(name: str, value: str = None, **options) -> "None|str":
        """Get response cookie for the given name if no value provided or add cookie to
        the response with the given name, value and options."""
        ...
    def delete_cookie(name: str) -> "Response":
        """Delete the cookie with the given name from the response."""
        ...
    def get_response_content(self) -> bytes:
        """Get response content."""
        ...
    def status(status: "str|int") -> "Response":
        """Set HTTP status code of the response."""
        ...
    def is_status(code: int) -> bool:
        """Check if response has the given status code."""
        ...
    def get_status_code(self) -> str:
        """Gets the HTTP status code of the response as a human string, like "200 OK"."""
        ...
    def get_status(self): ...
    def data(self) -> bytes:
        """Get the response content as bytes."""
        ...
    def converted_data(self) -> "str|bytes":
        """Get the response content as string or bytes so that the WSGI server handles it."""
        ...
    def view(view: Any, status: int = 200) -> "bytes|Response":
        """Set the response as a string or view."""
        ...
    def back(self) -> "Response":
        """Set the response as a redirect response back to previous path defined from the
        request."""
        ...
    def redirect(
        location: str = None,
        name: str = None,
        params: dict = {},
        url: str = None,
        status: int = 302,
    ) -> "Response":
        """Set the response as a redirect response. The redirection location can be defined
        with the location URL or with a route name. If a route name is used, route params can
        be provided."""

        ...
    def to_bytes(self) -> "bytes":
        """Converts the response to bytes."""
        ...
    def download(name: str, location: str, force: bool = False) -> "Response":
        """Set the response as a file download response."""
        ...



from tests import TestCase
from src.masonite.api.authentication import AuthenticatesTokens


class MockModel(AuthenticatesTokens):
    def save(self, *args, **kwargs):
        pass


class TestApiModule(TestCase):
    def test_can_generate_token(self):
        user = MockModel()

        self.assertTrue(len(user.generate_jwt()) == 195)



import os
import uuid
from os.path import isfile, join
from shutil import copyfile, move

from ..FileStream import FileStream
from ..File import File
from ...utils.filesystem import get_extension


class LocalDriver:
    def __init__(self, application):
        self.application = application
        self.options = {}

    def set_options(self, options):
        self.options = options
        return self

    def get_path(self, path):
        file_path = os.path.join(self.options.get("path"), path)
        self.make_file_path_if_not_exists(file_path)
        return file_path

    def get_name(self, path, alias):
        extension = get_extension(path)
        return f"{alias}{extension}"

    def put(self, file_path, content):
        with open(self.get_path(os.path.join(file_path)), "w") as f:
            f.write(content)
        return content

    def put_file(self, file_path, content, name=None):
        file_name = self.get_name(content.name, name or str(uuid.uuid4()))

        if hasattr(content, "get_content"):
            content = content.get_content()

        if isinstance(content, str):
            content = bytes(content, "utf-8")

        with open(self.get_path(os.path.join(file_path, file_name)), "wb") as f:
            f.write(content)

        return os.path.join(file_path, file_name)

    def get(self, file_path):
        try:
            with open(self.get_path(file_path), "r") as f:
                content = f.read()

            return content
        except FileNotFoundError:
            return None

    def exists(self, file_path):
        return os.path.exists(self.get_path(file_path))

    def missing(self, file_path):
        return not self.exists(file_path)

    def stream(self, file_path):
        with open(self.get_path(file_path), "r") as f:
            content = f
        return FileStream(content)

    def copy(self, from_file_path, to_file_path):
        return copyfile(from_file_path, to_file_path)

    def move(self, from_file_path, to_file_path):
        return move(self.get_path(from_file_path), self.get_path(to_file_path))

    def prepend(self, file_path, content):
        value = self.get(file_path)
        content = content + value
        self.put(file_path, content)
        return content

    def append(self, file_path, content):
        with open(self.get_path(file_path), "a") as f:
            f.write(content)
        return content

    def delete(self, file_path):
        return os.remove(self.get_path(file_path))

    def make_directory(self, directory):
        pass

    def store(self, file, name=None):
        if name:
            name = f"{name}{file.extension()}"
        full_path = self.get_path(name or file.hash_path_name())
        with open(full_path, "wb") as f:
            f.write(file.stream())

        return full_path

    def make_file_path_if_not_exists(self, file_path):
        if not os.path.isfile(file_path):
            if not os.path.exists(os.path.dirname(file_path)):
                # Create the path to the model if it does not exist
                os.makedirs(os.path.dirname(file_path))

            return True

        return False

    def get_files(self, directory=""):
        file_path = self.get_path(directory)
        files = []
        for f in os.listdir(file_path):
            if not isfile(join(file_path, f)):
                continue

            files.append(File(self.get(f), f))

        return files



"""Base Queue Module."""


class Queueable:
    """Makes classes Queueable."""

    run_again_on_fail = True
    run_times = 3

    def handle(self):
        pass

    def failed(self, obj, e):
        pass

    def __repr__(self):
        return self.__class__.__name__



from tests import TestCase

from src.masonite.facades import Dump
from src.masonite.exceptions import DumpException


class TestDumper(TestCase):
    def setUp(self):
        super().setUp()
        self.dumper = self.application.make("dumper")

    def tearDown(self):
        super().tearDown()
        self.dumper.clear()

    def test_get_dumps(self):
        dumps = self.dumper.get_dumps()
        assert dumps == []

    def test_get_serialized_dumps(self):
        dumps = self.dumper.get_serialized_dumps()
        assert dumps == []

    def test_dump(self):
        self.dumper.dump(1, {"test": "value"})
        dumps = self.dumper.get_dumps()
        assert len(dumps) == 1

        first_dump = self.dumper.last()
        assert first_dump.line == 25
        assert first_dump.filename.endswith("test_dumper.py")
        assert first_dump.method == "test_dump"
        assert len(first_dump.objects.keys()) == 2
        assert first_dump.objects.get("<class 'int'>") == 1
        assert first_dump.objects.get("<class 'dict'>") == {"test": "value"}

    def test_dump_can_get_variables_name(self):
        test = 1
        other_test = "a"
        self.dumper.dump(test, other_test)
        test_dump = self.dumper.last()
        assert test_dump.objects.get("test") == 1
        assert test_dump.objects.get("other_test") == "a"

    def test_serialize_dump(self):
        self.dumper.dump(1, {"test": "value"})
        data = self.dumper.last().serialize()
        assert data.get("line") == 46
        assert data.get("filename")
        assert data.get("timestamp")
        assert data.get("method")
        first_object = data.get("objects").get("<class 'int'>")
        assert first_object.get("value") == "1"

    def test_serialize_dump_properties(self):
        class TestObject:
            key = "value"
            _other_key = 1

        my_obj = TestObject()
        self.dumper.dump(my_obj)
        data = self.dumper.last().serialize()
        assert "my_obj" in data.get("objects")
        obj_props = data.get("objects").get("my_obj").get("properties")
        assert obj_props.get("private").get("_other_key") == "1"
        assert obj_props.get("public").get("key") == "value"

    def test_can_add_several_dumps(self):
        self.dumper.dump(1)
        self.dumper.dump(2)
        self.dumper.dump(3)
        assert len(self.dumper.get_dumps()) == 3

    def test_can_clear_dumps(self):
        self.dumper.dump(1)
        self.dumper.dump(2)
        self.dumper.clear()
        assert len(self.dumper.get_dumps()) == 0

    def test_dump_and_die(self):
        self.dumper.dump(1)
        with self.assertRaises(DumpException):
            self.dumper.dd(2)
        assert len(self.dumper.get_dumps()) == 2

    def test_dumps_are_ordered_by_most_recent(self):
        var = 1
        var_latest = 2
        self.dumper.dump(var)
        self.dumper.dump(var_latest)
        assert self.dumper.last().objects.get("var_latest") == 2
        assert self.dumper.get_dumps()[0].objects.get("var_latest") == 2

    def test_can_revert_dump_order(self):
        var = 1
        var_latest = 2
        self.dumper.dump(var)
        self.dumper.dump(var_latest)
        assert self.dumper.get_dumps(ascending=True)[0].objects.get("var") == 1

    def test_dump_facade(self):
        var = "test"
        Dump.dump(var)
        assert Dump.last().objects.get("var") == "test"

    def test_dump_builtins(self):
        var = "test"
        dump(var)
        assert self.dumper.last().objects.get("var") == "test"

    def test_dump_output_data_in_console(self):
        var = "test"
        dump(var)
        self.assertConsoleOutputContains(">>> DUMP")
        self.assertConsoleOutputContains("test")



import ipaddress
from .. import Middleware
from ...request import Request


class IpMiddleware(Middleware):

    # order of resolution of headers used to fetch request ip
    headers = [
        "HTTP_CLIENT_IP",
        "HTTP_X_FORWARDED_FOR",
        "HTTP_X_FORWARDED",
        "HTTP_X_CLUSTER_CLIENT_IP",
        "HTTP_FORWARDED_FOR",
        "HTTP_FORWARDED",
        "REMOTE_ADDR",
    ]

    def get_ip(self, request: Request):
        for header in self.headers:
            for raw_ip in request.environ.get(header, "").split(","):
                try:
                    ip = ipaddress.ip_address(raw_ip.strip())
                except ValueError:
                    continue
                if not ip.is_private and not ip.is_reserved:
                    return str(ip)
        return request.environ.get("REMOTE_ADDR")

    def before(self, request, response):
        request._ip = self.get_ip(request)
        return request

    def after(self, request, response):
        return request



import pathlib
import json
from tests import TestCase

from src.masonite.helpers import MixHelper
from src.masonite.exceptions import MixFileNotFound, MixManifestNotFound
from src.masonite.facades import Config


class TestMix(TestCase):
    def setUp(self):
        super().setUp()
        self.mix = MixHelper(self.application).url
        self.manifest_file = "mix-manifest.json"

    def tearDown(self):
        super().tearDown()
        try:
            # missing_ok=True is only available starting from Python3.8
            pathlib.Path(self.manifest_file).unlink()
        except FileNotFoundError:
            pass

    def _make_manifest(self):
        manifest = {"/storage/app.css": "/static/app.css"}
        with open("mix-manifest.json", "w") as f:
            json.dump(manifest, f)

    def test_mix_without_manifest_raise_exception(self):
        with self.assertRaises(MixManifestNotFound):
            self.mix("/storage/app.css")

    def test_mix_with_wrong_path_raise_exception(self):
        self._make_manifest()
        with self.assertRaises(MixFileNotFound):
            self.mix("/storage/wrong.css")

    def test_mix_url(self):
        self._make_manifest()
        self.assertEqual(
            self.mix("/storage/app.css"), "http://localhost:8000/static/app.css"
        )
        # also works if missing / prefix
        self.assertEqual(
            self.mix("storage/app.css"), "http://localhost:8000/static/app.css"
        )

    def test_mix_with_mix_base_url(self):
        Config.set("application.mix_base_url", "https://some-cdn.com/")
        self._make_manifest()
        self.assertEqual(
            self.mix("storage/app.css"), "https://some-cdn.com/static/app.css"
        )
        Config.set("application.mix_base_url", None)



from tests import TestCase

from src.masonite.utils.str import (
    random_string,
    removeprefix,
    removesuffix,
    get_controller_name,
)

from tests.integrations.controllers.api.TestController import TestController


class TestStringsUtils(TestCase):
    def test_random_string(self):
        self.assertEqual(len(random_string()), 4)
        self.assertEqual(len(random_string(10)), 10)
        self.assertIsInstance(random_string(5), str)
        self.assertNotEqual(random_string(), random_string())

    def test_removesuffix(self):
        self.assertEqual(removesuffix("test.com", ".com"), "test")
        self.assertEqual(removesuffix("test", ".com"), "test")

    def test_removeprefix(self):
        self.assertEqual(removeprefix("AppEvent", "App"), "Event")
        self.assertEqual(removeprefix("Event", "App"), "Event")

    def test_get_controller_name(self):
        self.assertEqual(
            get_controller_name("WelcomeController@show"), "WelcomeController@show"
        )
        self.assertEqual(get_controller_name(TestController), "TestController@__call__")
        self.assertEqual(
            get_controller_name(TestController.show), "TestController@show"
        )
        controller = TestController()
        self.assertEqual(get_controller_name(controller), "TestController@__call__")



class HasColoredOutput:
    """Add level-colored output print functions to a class."""

    def success(self, message):
        print("\033[92m {0} \033[0m".format(message))

    def warning(self, message):
        print("\033[93m {0} \033[0m".format(message))

    def danger(self, message):
        print("\033[91m {0} \033[0m".format(message))

    def info(self, message):
        return self.success(message)


class AddCommandColors:
    """The default style set used by Cleo is defined here:
    https://github.com/sdispater/clikit/blob/master/src/clikit/formatter/default_style_set.py
    This mixin add method helper to output errors and warnings.
    """

    def error(self, text):
        """
        Write a string as information output.

        :param text: The line to write
        :type text: str
        """
        self.line(text, "error")

    def warning(self, text):
        """
        Write a string as information output.

        :param text: The line to write
        :type text: str
        """
        self.line(text, "c2")



class DatabaseTransactions:
    def startTestRun(self):
        self.application.make("resolver").begin_transaction(self.connection)
        return self

    def stopTestRun(self):
        self.application.make("resolver").rollback(self.connection)
        return self



import pendulum
import uuid
from ..routes import Route


class Auth:
    def __init__(self, application, guard_config=None):
        self.application = application
        self.guards = {}
        self._guard = None
        self.guard_config = guard_config or {}
        self.options = {}

        self._user = None

    def add_guard(self, name, guard):
        self.guards.update({name: guard})

    def set_configuration(self, config):
        self.guard_config = config
        return self

    def guard(self, guard):
        self._guard = guard
        return self

    def get_guard(self, name=None):
        if name is None and self._guard is None:
            return self.guards[self.guard_config.get("default")]

        return self.guards[self._guard]

    def get_config_options(self, guard=None):
        if guard is None:
            options = self.guard_config.get(self.guard_config.get("default"), {})
            options.update(self.options)
            return options

        options = self.guard_config.get(guard, {})
        options.update(self.options)
        return options

    def attempt(self, email, password, once=False):
        auth_config = self.get_config_options()
        auth_config.update({"once": once})
        user = self.get_guard().set_options(auth_config).attempt(email, password)
        self.set_user(user)
        return user

    def attempt_by_id(self, user_id, once=False):
        auth_config = self.get_config_options()
        auth_config.update({"once": once})
        user = self.get_guard().set_options(auth_config).attempt_by_id(user_id)
        self.set_user(user)
        return user

    def logout(self):
        """Logout the current authenticated user.

        Returns:
            self
        """
        self.application.make("request").remove_user()
        self.application.make("response").delete_cookie("token")
        self.get_guard().set_options(self.get_config_options()).logout()
        self._user = None

    def user(self):
        """Get the current authenticated user.

        Returns:
            self
        """
        if self._user:
            return self._user
        return self.get_guard().set_options(self.get_config_options()).user()

    def register(self, dictionary):
        """Logout the current authenticated user.

        Returns:
            self
        """
        auth_config = self.get_config_options()
        return self.get_guard().set_options(auth_config).register(dictionary)

    def set_user(self, user):
        self._user = user
        return self

    def password_reset(self, email):
        """Logout the current authenticated user.

        Returns:
            self
        """
        token = str(uuid.uuid4())
        auth_config = self.get_config_options()
        auth = self.get_guard().set_options(auth_config).get_auth_column(email)
        if not auth:
            return (None, None)
        try:
            existing = (
                self.application.make("builder")
                .new()
                .table(self.guard_config.get("password_reset_table"))
                .where("email", email)
                .first()
            )
            if existing:
                token = existing["token"]
            else:
                self.application.make("builder").new().table(
                    self.guard_config.get("password_reset_table")
                ).create(
                    {
                        "email": email,
                        "token": token,
                        "expires_at": pendulum.now()
                        .add(minutes=self.guard_config.get("password_reset_expiration"))
                        .to_datetime_string()
                        if self.guard_config.get("password_reset_expiration")
                        else None,
                        "created_at": pendulum.now().to_datetime_string(),
                    }
                )
        except Exception:
            return (None, None)

        self.application.make("event").fire("auth.password_reset", email, token)
        return auth, token

    def reset_password(self, password, token):
        """Logout the current authenticated user.

        Returns:
            self
        """

        reset_record = (
            self.application.make("builder")
            .new()
            .table(self.guard_config.get("password_reset_table"))
            .where("token", token)
            .first()
        )

        auth_config = self.get_config_options()
        (
            self.get_guard()
            .set_options(auth_config)
            .reset_password(reset_record.get("email"), password)
        )

        (
            self.application.make("builder")
            .new()
            .table(self.guard_config.get("password_reset_table"))
            .where("token", token)
            .delete()
        )

        return True

    @classmethod
    def routes(self):
        return [
            Route.get("/login", "auth.LoginController@show").name("login"),
            Route.get("/logout", "auth.LoginController@logout").name("logout"),
            Route.get("/home", "auth.HomeController@show")
            .name("auth.home")
            .middleware("auth"),
            Route.get("/register", "auth.RegisterController@show").name("register"),
            Route.post("/register", "auth.RegisterController@store").name(
                "register.store"
            ),
            Route.get("/password_reset", "auth.PasswordResetController@show").name(
                "password_reset"
            ),
            Route.post("/password_reset", "auth.PasswordResetController@store").name(
                "password_reset.store"
            ),
            Route.get(
                "/change_password/@token",
                "auth.PasswordResetController@change_password",
            ).name("change_password"),
            Route.post(
                "/change_password/@token",
                "auth.PasswordResetController@store_changed_password",
            ).name("change_password.store"),
            Route.post("/login", "auth.LoginController@store").name("login.store"),
        ]



from ..facades import Loader
from ..utils.structures import data
from ..exceptions import InvalidConfigurationLocation, InvalidConfigurationSetup


class Configuration:
    # Foundation configuration keys
    reserved_keys = [
        "application",
        "auth",
        "broadcast",
        "cache",
        "database",
        "filesystem",
        "mail",
        "notification",
        "providers",
        "queue",
        "session",
    ]

    def __init__(self, application):
        self.application = application
        self._config = data()

    def load(self):
        """At boot load configuration from all files and store them in here."""
        config_root = self.application.make("config.location")
        for module_name, module in Loader.get_modules(
            config_root, raise_exception=True
        ).items():
            params = Loader.get_parameters(module)
            for name, value in params.items():
                self._config[f"{module_name}.{name.lower()}"] = value

        # check loaded configuration
        if not self._config.get("application"):
            raise InvalidConfigurationLocation(
                f"Config directory {config_root} does not contain required configuration files."
            )

    def merge_with(self, path, external_config):
        """Merge external config at key with project config at same key. It's especially
        useful in Masonite packages in order to merge the configuration default package with
        the package configuration which can be published in project.

        This functions disallow merging configuration using foundation configuration keys
        (such as 'application').
        """
        if path in self.reserved_keys:
            raise InvalidConfigurationSetup(
                f"{path} is a reserved configuration key name. Please use an other key."
            )
        if isinstance(external_config, str):
            # config is a path and should be loaded
            params = Loader.get_parameters(external_config)
        else:
            params = external_config
        base_config = {name.lower(): value for name, value in params.items()}
        merged_config = {**base_config, **self.get(path, {})}
        self.set(path, merged_config)

    def set(self, path, value):
        self._config[path] = value

    def has(self, path):
        return path in self._config

    def all(self):
        return self._config

    def get(self, path, default=None):
        try:
            config_at_path = self._config[path]
            if isinstance(config_at_path, dict):
                return data(config_at_path)
            else:
                return config_at_path
        except KeyError:
            return default



import time

import pytest

from src.masonite.events import Event
from tests import TestCase


class UserAddedEvent(Event):
    def __init__(self):
        pass

    def handle(self):
        pass


class NewUserEvent(Event):
    def __init__(self):
        pass

    def handle(self):
        pass


class SendEmailListener:
    def handle(self, event):
        pass


class UpdateAdminListener:
    def handle(self, event, user):
        pass


class SendAlert:
    def handle(self, event):
        pass


class AdminNotificationListener:
    def handle(self):
        pass


class Subscriber:
    def handle(self, event):
        pass

    def subscribe(self, event):
        event.listen("masonite.event_handled", [self.__class__])


class TestEvent(TestCase):
    def setUp(self):
        super().setUp()
        self.event = self.application.make("event")
        self.event.listen(UserAddedEvent, [SendEmailListener])

    def test_events_registered(self):
        self.assertEqual(len(self.event.get_events().get(UserAddedEvent)), 2)
        self.event.listen(UserAddedEvent, [AdminNotificationListener])
        self.assertEqual(len(self.event.get_events().get(UserAddedEvent)), 3)

    def test_fire_event_class(self):
        self.event.fire(UserAddedEvent)

    def test_fire_event_string(self):
        self.event.listen("masonite.*.booted", [SendEmailListener])
        self.event.listen("masonite.commands", [SendEmailListener])
        self.event.listen("view.*", [SendEmailListener])
        self.event.listen("masonite.exception.*", [SendEmailListener])
        self.event.listen("user.added", [UpdateAdminListener])
        self.assertEqual(self.event.fire("masonite.commands"), ["masonite.commands"])
        self.assertEqual(self.event.fire("masonite.orm.booted"), ["masonite.*.booted"])
        self.assertEqual(self.event.fire("masonite.orm"), [])
        self.assertEqual(self.event.fire("masonite.command"), [])
        self.assertEqual(
            self.event.fire("masonite.exception.ZeroDivisionError"),
            ["masonite.exception.*"],
        )
        self.assertEqual(self.event.fire("view.rendered"), ["view.*"])
        self.assertEqual(self.event.fire("user.added", 1), ["user.added"])

    def test_fire_event_class(self):
        self.event.listen(NewUserEvent, [SendAlert])
        self.event.fire(NewUserEvent())

    def test_can_subscribe(self):
        self.event.subscribe(Subscriber())
        self.assertEqual(
            self.event.fire("masonite.event_handled"), ["masonite.event_handled"]
        )



from tests import TestCase
from masoniteorm.models import Model

from src.masonite.exceptions.exceptions import AuthorizationException, GateDoesNotExist
from src.masonite.authorization import AuthorizationResponse
from src.masonite.routes import Route


class User(Model):
    """User Model"""

    __fillable__ = ["name", "email", "password"]


class Post(Model):
    __fillable__ = ["user_id", "name"]


class TestGate(TestCase):
    def setUp(self):
        super().setUp()
        self.gate = self.application.make("gate")
        self.make_request()
        self.make_response()
        self.addRoutes(
            Route.get("/not-authorized", "WelcomeController@not_authorized"),
            Route.get(
                "/authorize-helper", "WelcomeController@use_authorization_helper"
            ),
            Route.get("/authorizations", "WelcomeController@authorizations"),
        )

    def tearDown(self):
        super().tearDown()
        self.gate.permissions = {}
        self.gate.before_callbacks = []
        self.gate.after_callbacks = []

    def test_can_define_gates(self):
        self.gate.define(
            "create-post", lambda user: user.email == "idmann509@gmail.com"
        )
        self.gate.define("view-post", lambda user: True)
        self.assertEqual(
            list(self.gate.permissions.keys()), ["create-post", "view-post"]
        )

    def test_can_use_for_user(self):
        given_user = User()
        user = self.gate.for_user(given_user)._get_user()
        self.assertEqual(user, given_user)

    def test_gate_user_the_current_authenticated_request_user(self):
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        user = self.gate._get_user()
        self.assertEqual(user.email, "idmann509@gmail.com")

    def test_can_use_gate_facade(self):
        from src.masonite.facades import Gate

        self.assertEqual(self.gate.define, Gate.define)

    def test_denies_guests_users(self):
        self.gate.define("create-post", lambda user: True)
        # above permission should always been allowed but user is not authenticated
        self.assertTrue(self.gate.denies("create-post"))

    def test_allows_guests_users_if_specified(self):
        # it can be specified by allowing user to be optional
        self.gate.define("create-post", lambda user=None: True)
        # above permission should always been allowed even for guests users
        self.assertTrue(self.gate.allows("create-post"))

    def test_allows_and_denies(self):
        self.gate.define(
            "create-post", lambda user: user.email == "idmann509@gmail.com"
        )
        self.gate.define("view-post", lambda user: False)
        # authenticate user
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")

        self.assertTrue(self.gate.allows("create-post"))
        self.assertFalse(self.gate.denies("create-post"))
        self.assertFalse(self.gate.allows("view-post"))
        self.assertTrue(self.gate.denies("view-post"))

    def test_allows_with_arg(self):
        self.gate.define("update-post", lambda user, post: post.user_id == user.id)
        # authenticate user
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        # create a post for the user 1
        post = Post()
        post.user_id = 1
        post2 = Post()
        post2.user_id = 3
        self.assertTrue(self.gate.allows("update-post", post))
        self.assertFalse(self.gate.allows("update-post", post2))

    def test_gate_has_permission(self):
        self.gate.define("display-admin", lambda user: False)
        self.assertTrue(self.gate.has("display-admin"))
        self.assertFalse(self.gate.has("view-user"))

    def test_authorize(self):
        self.withExceptionsHandling()
        self.gate.define("display-admin", lambda user: False)
        with self.debugMode():
            with self.assertRaises(AuthorizationException) as e:
                self.gate.authorize("display-admin")
                exception = e.exception
            self.assertEqual(e.exception.message, "Action not authorized")
            self.assertEqual(e.exception.status, 403)

    def test_authorize_in_controller(self):
        self.withExceptionsHandling()  # this will allow exception to be handled and rendered
        self.gate.define("display-admin", lambda user: False)
        with self.debugMode(False):
            self.get("/not-authorized").assertForbidden().assertContains(
                "Permission Denied"
            )

    def test_inspect(self):
        self.gate.define("display-admin", lambda user: False)
        response = self.gate.inspect("display-admin")
        self.assertIsInstance(response, AuthorizationResponse)
        self.assertFalse(response.allowed())

    def test_define_gate_returning_response(self):
        self.gate.define(
            "display-admin",
            lambda user: AuthorizationResponse.allow()
            if user.email == "admin@masonite.com"
            else AuthorizationResponse.deny("You shall not pass"),
        )
        # authenticate user
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        response = self.gate.inspect("display-admin")
        self.assertFalse(response.allowed())
        self.assertEqual(response.message(), "You shall not pass")

    def test_gate_before(self):
        self.gate.before(lambda user, permission: user.email == "idmann509@gmail.com")
        # a permission that is always False
        self.gate.define("display-admin", lambda user: False)
        # authenticate user
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        self.assertTrue(self.gate.allows("display-admin"))

    def test_gate_after(self):
        self.gate.after(lambda user, permission, result: False)
        # a permission that is always False
        self.gate.define("display-admin", lambda user: True)
        # authenticate user 1
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        self.assertTrue(self.gate.denies("display-admin"))

    def test_any(self):
        self.gate.define("delete-post", lambda user, post: True)
        self.gate.define("update-post", lambda user, post: user.id == post.user_id)
        # authenticate user 1
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        post = Post()
        post.user_id = 1
        self.assertTrue(self.gate.any(["update-post", "delete-post"], post))

    def test_none(self):
        self.gate.define("force-delete-post", lambda user, post: False)
        self.gate.define(
            "restore-post", lambda user, post: user.email == "admin@gmail.com"
        )
        # authenticate user 1
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        post = Post()
        self.assertTrue(self.gate.none(["force-delete-post", "restore-post"], post))

    def test_unknown_gate_raises_exception(self):
        with self.assertRaises(GateDoesNotExist):
            self.gate.allows("can-fly")

    def test_view_helpers(self):
        from tests.integrations.app.User import User

        self.gate.define("view-posts", lambda user=None: True)
        self.gate.define(
            "display-admin", lambda user: user.email == "idmann509@gmail.com"
        )

        self.actingAs(User.find(1)).get("/authorizations").assertContains(
            "User can display admin"
        ).assertContains("User can view posts")

        self.actingAsGuest().get("/authorizations").assertContains(
            "User can view posts"
        ).assertNotContains("User can display admin")

    def test_can_use_authorize_helper_on_request(self):
        from tests.integrations.app.User import User

        self.gate.define("view-posts", lambda user=None: True)
        self.gate.define(
            "display-admin", lambda user: user.email == "idmann509@gmail.com"
        )

        self.actingAs(User.find(1)).get("/authorize-helper").assertOk()



from inspect import isclass, signature

from .AuthorizationResponse import AuthorizationResponse
from ..exceptions.exceptions import GateDoesNotExist, PolicyDoesNotExist


class Gate:
    def __init__(
        self,
        application,
        user_callback=None,
        policies={},
        permissions={},
        before_callbacks=[],
        after_callbacks=[],
    ):
        self.application = application
        self.user_callback = user_callback

        self.policies = policies
        self.permissions = permissions
        self.before_callbacks = before_callbacks
        self.after_callbacks = after_callbacks

    def define(self, permission, condition):
        if not callable(condition):
            raise Exception(f"Permission {permission} should be given a callable.")

        self.permissions.update({permission: condition})

    def register_policies(self, policies):
        for model_class, policy_class in policies:
            self.policies[model_class] = policy_class
        return self

    def get_policy_for(self, instance_or_class):
        if isinstance(instance_or_class, str):
            # TODO: load model from str, get class and get policies
            policy = None
        elif isclass(instance_or_class):
            policy = self.policies.get(instance_or_class, None)
        else:
            policy = self.policies.get(instance_or_class.__class__, None)
        if policy:
            return policy()
        else:
            return None

    def before(self, before_callback):
        if not callable(before_callback):
            raise Exception("before() should be given a callable.")
        self.before_callbacks.append(before_callback)

    def after(self, after_callback):
        if not callable(after_callback):
            raise Exception("before() should be given a callable.")
        self.after_callbacks.append(after_callback)

    def allows(self, permission, *args):
        return self.inspect(permission, *args).allowed()

    def denies(self, permission, *args):
        return not self.inspect(permission, *args).allowed()

    def has(self, permission):
        return permission in self.permissions

    def for_user(self, user):
        return Gate(
            self.application,
            lambda: user,
            self.policies,
            self.permissions,
            self.before_callbacks,
            self.after_callbacks,
        )

    def any(self, permissions, *args):
        """Check that every of those permissions are allowed."""
        for permission in permissions:
            if self.denies(permission, *args):
                return False
        return True

    def none(self, permissions, *args):
        """Check that none of those permissions are allowed."""
        for permission in permissions:
            if self.allows(permission, *args):
                return False
        return True

    def authorize(self, permission, *args):
        return self.inspect(permission, *args).authorize()

    def inspect(self, permission, *args):
        """Get permission checks results for the given user then builds and returns an
        authorization response."""
        boolean_result = self.check(permission, *args)
        if isinstance(boolean_result, AuthorizationResponse):
            return boolean_result
        if boolean_result:
            return AuthorizationResponse.allow()
        else:
            return AuthorizationResponse.deny()

    def check(self, permission, *args):
        """The core of the authorization class. Run before() checks, permission check and then
        after() checks."""
        user = self._get_user()

        # run before checks and returns immediately if non null response
        result = None
        for callback in self.before_callbacks:
            result = callback(user, permission)
            if result:
                break

        # run permission checks if nothing returned previously
        if result is None:
            # first check in policy
            permission_method = None
            if len(args) > 0:
                policy = self.get_policy_for(args[0])
                if policy:
                    try:
                        permission_method = getattr(policy, permission)
                    except AttributeError:
                        raise PolicyDoesNotExist(
                            f"Policy method {permission} not found in {policy.__class__.__name__}."
                        )

            if not permission_method:
                # else check in gates
                try:
                    permission_method = self.permissions[permission]
                except KeyError:
                    raise GateDoesNotExist(
                        f"Gate {permission} has not been found. Did you declare it ?"
                    )

            params = signature(permission_method).parameters
            # check if user parameter is optional (meaning that guests users are allowed)
            if (
                permission_method.__defaults__
                and permission_method.__defaults__[0] is None
                and not user
            ):
                result = True
            elif not user:
                result = False
            elif len(params) == 1:
                result = permission_method(user)
            else:
                result = permission_method(user, *args)

        # run after checks
        for callback in self.after_callbacks:
            after_result = callback(user, permission, result)
            result = after_result if after_result is not None else result

        return result

    def _get_user(self):
        from ..facades import Request

        if self.user_callback:
            return self.user_callback()
        else:
            return Request.user()



import pendulum

from masoniteorm.models import Model

from tests import TestCase
from src.masonite.tests import DatabaseTransactions
from src.masonite.notification import Notification, Notifiable
from src.masonite.notification import DatabaseNotification


class User(Model, Notifiable):
    """User Model"""

    __fillable__ = ["name", "email", "password"]


class WelcomeNotification(Notification):
    def to_database(self, notifiable):
        return {"data": "Welcome {0}!".format(notifiable.name)}

    def via(self, notifiable):
        return ["database"]


notif_data = {
    "id": "1234",
    "read_at": None,
    "type": "TestNotification",
    "data": "test",
    "notifiable_id": 1,
    "notifiable_type": "users",
}


class TestDatabaseDriver(TestCase, DatabaseTransactions):
    connection = None

    def setUp(self):
        super().setUp()
        self.notification = self.application.make("notification")

    def test_send_to_notifiable(self):
        user = User.find(1)
        count = user.notifications.count()
        user.notify(WelcomeNotification())
        assert user.notifications.count() == count + 1

    def test_database_notification_is_created_correctly(self):
        user = User.find(1)
        notification = user.notify(WelcomeNotification())
        assert notification["id"]
        assert not notification["read_at"]
        assert notification["data"] == '{"data": "Welcome Joe!"}'
        assert notification["notifiable_id"] == user.id
        assert notification["notifiable_type"] == "users"

    def test_notify_multiple_users(self):
        User.create({"name": "sam", "email": "sam@test.com", "password": "secret"})
        users = User.all()  # == 2 users
        self.notification.send(users, WelcomeNotification())
        assert users[0].notifications.count() == 1
        assert users[1].notifications.count() == 1


class TestDatabaseNotification(TestCase, DatabaseTransactions):
    connection = None

    def test_database_notification_read_state(self):
        notification = DatabaseNotification.create(
            {
                **notif_data,
                "read_at": pendulum.now().to_datetime_string(),
            }
        )
        self.assertTrue(notification.is_read)
        notification.read_at = None
        self.assertFalse(notification.is_read)

    def test_database_notification_unread_state(self):
        notification = DatabaseNotification.create(
            {
                **notif_data,
                "read_at": pendulum.yesterday().to_datetime_string(),
            }
        )
        self.assertFalse(notification.is_unread)
        notification.read_at = None
        self.assertTrue(notification.is_unread)

    def test_database_notification_mark_as_read(self):
        notification = DatabaseNotification.create(notif_data)
        notification.mark_as_read()
        self.assertNotEqual(None, notification.read_at)

    def test_database_notification_mark_as_unread(self):
        notification = DatabaseNotification.create(
            {
                **notif_data,
                "read_at": pendulum.now().to_datetime_string(),
            }
        )
        notification.mark_as_unread()
        self.assertEqual(None, notification.read_at)

    def test_notifiable_get_notifications(self):
        user = User.find(1)
        self.assertEqual(0, user.notifications.count())
        user.notify(WelcomeNotification())
        self.assertEqual(1, user.notifications.count())

    def test_notifiable_get_read_notifications(self):
        user = User.find(1)
        self.assertEqual(0, user.read_notifications.count())
        DatabaseNotification.create(
            {
                **notif_data,
                "read_at": pendulum.yesterday().to_datetime_string(),
                "notifiable_id": user.id,
            }
        )
        self.assertEqual(1, user.read_notifications.count())

    def test_notifiable_get_unread_notifications(self):
        user = User.find(1)
        self.assertEqual(0, user.unread_notifications.count())
        DatabaseNotification.create(
            {
                **notif_data,
                "notifiable_id": user.id,
            }
        )
        self.assertEqual(1, user.unread_notifications.count())



class CorsMiddleware:

    headers = {}

    def before(self, request, response):
        if request.get_request_method() == "OPTIONS":
            return response.view("preflight")
        return request

    def after(self, request, response):
        response.header(self.headers)
        return request



class Url:
    """URL helper facade."""

    def url(path: str = "") -> str:
        """Generates a fully qualified url to the given path. If no path is given this will return
        the base url domain."""
        ...
    def asset(alias: str, filename: str) -> str:
        """Generates a fully qualified URL for the given asset using the given disk
        Example:
            asset("local", "avatar.jpg") (take first pat)
            asset("s3.private", "doc.pdf") (when multiple paths are specified for the disk)
        """
        ...
    def route(name: str, params: dict = {}, absolute: bool = True) -> str:
        """Generates a fully qualified URL to the given route name.
        Example:
            route("users.home") : http://masonite.app/dashboard/
            route("users.profile", {"id": 1}) : http://masonite.app/users/1/profile/
            route("users.profile", {"id": 1}, absolute=False) : /users/1/profile/
        """
        ...



from tests import TestCase

from src.masonite.helpers import url


class TestUrlsHelper(TestCase):
    def test_url(self):
        self.assertEqual(url.url("about/us"), "http://localhost:8000/about/us")
        self.assertEqual(url.url("/about/us"), "http://localhost:8000/about/us")
        self.assertEqual(url.url(), "http://localhost:8000/")

        # with query parameters
        self.assertEqual(
            url.url("search", {"q": "joe", "order": "asc"}),
            "http://localhost:8000/search?q=joe&order=asc",
        )
        self.assertEqual(
            url.url("search", {"q": "El Niño", "order": "asc"}),
            "http://localhost:8000/search?q=El+Ni%C3%B1o&order=asc",
        )

        self.assertEqual(
            url.url("search?q=joe", {"order": "asc"}),
            "http://localhost:8000/search?q=joe&order=asc",
        )

        self.assertEqual(
            url.url("/users/search?q=joe", {"q": "john doe", "debug": 1}),
            "http://localhost:8000/users/search?q=john+doe&debug=1",
        )

    def test_route(self):
        self.assertEqual(url.route("welcome"), "http://localhost:8000/")
        self.assertEqual(
            url.route("users.profile", {"id": 1}), "http://localhost:8000/users/1"
        )
        self.assertEqual(url.route("upload"), "http://localhost:8000/upload")
        self.assertEqual(url.route("upload", absolute=False), "/upload")
        # with query parameters
        self.assertEqual(
            url.route("upload", query_params={"force": 1}),
            "http://localhost:8000/upload?force=1",
        )

    def test_asset(self):
        self.assertTrue(
            url.asset("local", "test.jpg").endswith(
                "storage/framework/filesystem/test.jpg"
            )
        )



from tests import TestCase

from src.masonite.helpers import optional


class SomeClass:

    my_attr = 3

    def my_method(self):
        return 4


class TestOptionalHelper(TestCase):
    def test_optional_with_existing(self):
        obj = SomeClass()
        assert optional(obj).my_attr == 3
        assert optional(obj).my_method() == 4

    def test_optional_with_undefined(self):
        obj = SomeClass()
        assert optional(obj).non_existing_attr is None

        # not beautiful but we can do this to handle calling methods
        assert optional(optional(obj).non_existing_method)() is None

    def test_optional_with_undefined_on_none(self):
        obj = None
        assert optional(obj).non_existing_attr is None

    def test_optional_with_default(self):
        obj = SomeClass()
        assert optional(obj, 0).non_existing_attr == 0

    def test_optional_with_callable_default(self):
        obj = SomeClass()
        assert optional(obj, lambda the_obj: "a").non_existing_attr == "a"



import time
import inspect
from pprint import pformat


def color(string):
    return f"\033[93m{string}\033[0m"


def is_property(obj):
    return not inspect.ismethod(obj)


def is_local(obj_name, obj):
    return (
        not obj_name.startswith("__")
        and not obj_name.endswith("__")
        and not type(obj).__name__ == "builtin_function_or_method"
    )


def is_private(obj_name):
    return obj_name.startswith("_")


def serialize_property(obj):
    if isinstance(obj, list):
        local_list = []
        for subobj in obj:
            local_list.append(serialize_property(subobj))
        return local_list
    elif isinstance(obj, dict):
        local_dict = {}
        for key, val in obj.items():
            local_dict.update({key: serialize_property(val)})
        return local_dict
    elif hasattr(obj, "serialize"):
        return obj.serialize()
    else:
        return str(obj)


class Dump:
    def __init__(self, objects, method, filename, line):
        self.objects = objects
        self.method = method
        self.filename = filename
        self.line = line
        self.timestamp = time.time()

    def serialize(self):
        objects = {}
        for name, obj in self.objects.items():
            # serialize all obj properties
            all_properties = inspect.getmembers(obj, predicate=is_property)
            local_properties = {"private": {}, "public": {}}
            for prop_name, prop in all_properties:
                if is_local(prop_name, prop):
                    entry = {prop_name: serialize_property(prop)}
                    if is_private(prop_name):
                        local_properties["private"].update(entry)
                    else:
                        local_properties["public"].update(entry)

            objects.update({name: {"value": str(obj), "properties": local_properties}})

        return {
            "objects": objects,
            "method": self.method,
            "filename": self.filename,
            "line": self.line,
            "timestamp": self.timestamp,
        }

    def __repr__(self):
        return self._format()

    def __str__(self):
        return self._format()

    def _format(self):
        """Format the dump as string to be printed in console."""
        output = f"\n{color('>>> DUMP')} from {self.filename}: {color(f'L{self.line}')} in {color(f'{self.method}()')}"

        for name, obj in self.objects.items():
            output += f"\n\n{color(f'  - {name}:')}\n"
            output += f"  {pformat(obj, width=110, indent=4)}"

        output += color("\n\n<<< END")
        return output



from tests.features.notification.test_anonymous_notifiable import WelcomeNotification
from tests import TestCase
from src.masonite.mail import Mailable
from src.masonite.notification import Notification, SlackMessage, Sms, Notifiable
from masoniteorm.models import Model


class User(Model, Notifiable):
    """User Model"""

    __fillable__ = ["name", "email", "password"]

    def route_notification_for_broadcast(self):
        return f"user.{self.id}"

    def route_notification_for_slack(self):
        return "#bot"

    def route_notification_for_vonage(self):
        return "+33123456789"

    def route_notification_for_mail(self):
        return self.email


class WelcomeNotification(Notification):
    def to_mail(self, notifiable):
        return (
            Mailable()
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
        )

    def to_database(self, notifiable):
        return {"data": "Welcome {0}!".format(notifiable.name)}

    def to_broadcast(self, notifiable):
        return {"data": "Welcome"}

    def to_slack(self, notifiable):
        return SlackMessage().text("Welcome !").from_("test-bot")

    def to_vonage(self, notifiable):
        return Sms().text("Welcome !").from_("123456")

    def via(self, notifiable):
        return ["mail", "database", "slack", "vonage", "broadcast"]


class TestIntegrationsNotifications(TestCase):
    def setUp(self):
        super().setUp()
        self.fake("notification")

    def tearDown(self):
        super().tearDown()
        self.restore("notification")

    def test_all_drivers_with_anonymous(self):
        notif = (
            self.application.make("notification")
            .route("mail", "user@example.com")
            .route("slack", "#general")
            .route("broadcast", "all")
            .route("vonage", "+33456789012")
        )
        notif.send(WelcomeNotification()).assertSentTo(
            "user@example.com", WelcomeNotification
        ).assertSentTo("#general", WelcomeNotification).assertSentTo(
            "all", WelcomeNotification
        ).assertSentTo(
            "+33456789012", WelcomeNotification
        )

    def test_all_drivers_with_notifiable(self):
        self.application.make("notification").assertNothingSent()
        user = User.find(1)
        user.notify(WelcomeNotification())
        self.application.make("notification").assertSentTo(
            "user@example.com", WelcomeNotification
        ).assertSentTo("#general", WelcomeNotification).assertSentTo(
            "all", WelcomeNotification
        ).assertSentTo(
            "+33456789012", WelcomeNotification
        )



from tests import TestCase
from tests.integrations.app.User import User
from src.masonite.tests import DatabaseTransactions


class TestDatabase(TestCase, DatabaseTransactions):

    connection = None

    def setUp(self):
        super().setUp()

    def test_can_use_transactions(self):
        User.create({"name": "john", "email": "john6", "password": "secret"})

    # def test_assert_deleted(self):
    #     user = User.find(1)
    #     user.delete()
    #     self.assertDeleted(user)



from tests import TestCase

from src.masonite.notification import Notification, AnonymousNotifiable
from src.masonite.mail import Mailable


class WelcomeNotification(Notification):
    def to_mail(self, notifiable):
        return Mailable().text("Welcome")

    def via(self, notifiable):
        return ["mail"]


class TestAnonymousNotifiable(TestCase):
    def test_one_routing(self):
        notifiable = AnonymousNotifiable(self.application).route(
            "mail", "user@example.com"
        )
        self.assertDictEqual({"mail": "user@example.com"}, notifiable._routes)

    def test_multiple_routing(self):
        notifiable = (
            AnonymousNotifiable(self.application)
            .route("mail", "user@example.com")
            .route("slack", "#general")
        )
        self.assertDictEqual(
            {"mail": "user@example.com", "slack": "#general"}, notifiable._routes
        )

    def test_sending_notification(self):
        self.application.make("notification").route("mail", "user@example.com").send(
            WelcomeNotification()
        )

    def test_can_override_dry_when_sending(self):
        AnonymousNotifiable(self.application).route("mail", "user@example.com").send(
            WelcomeNotification(), dry=True
        )
        self.application.make("notification").dry_notifications.keys() == 1

    def test_can_override_fail_silently_when_sending(self):
        class FailingNotification(Notification):
            def to_slack(self, notifiable):
                raise Exception("Error")

            def via(self, notifiable):
                return ["slack"]

        AnonymousNotifiable(self.application).route("slack", "#general").send(
            FailingNotification(), fail_silently=True
        )
        # no assertion raised :)



"""The Message Bag Module"""

import json


class MessageBag:
    def __init__(self, items={}):
        self.items = items

    def add(self, error, message):
        """Adds an error and message to the message bag

        Arguments:
            error {string} -- The error to add
            message {string} -- The message to add
        """

        if error in self.items:
            self.items[error].append(message)
        else:
            self.items.update({error: [message]})

    def all(self):
        """Get all errors and messages"""
        return self.items

    def any(self):
        """If the messagebag has any errors"""
        return len(self.items) > 0

    def has(self, key):
        """If the messagebag has any errors"""
        return key in self.all()

    def empty(self):
        """If the messagebag has any errors"""
        return not self.any()

    def first(self, key):
        """Gets the first error and message"""
        return self.get(key)[0]

    def count(self):
        """Gets the amount of errors"""
        return len(self.items)

    def json(self):
        """Gets the amount of errors"""
        return json.dumps(self.items)

    def amount(self, key):
        """Gets the amount of messages

        Arguments:
            key {string} -- the error to get the amount of.

        Returns:
            int -- Returns the amount of messages
        """
        return len(self.items[key])

    def get(self, key):
        """Gets all the messages for a specific error.

        Arguments:
            key {string} -- the error to get the messages for

        Returns:
            list -- list of errors
        """
        return self.items[key]

    def errors(self):
        """Gets a list of errors"""
        return list(self.items.keys())

    def messages(self):
        """Gets a list of all the messages"""
        messages = []
        for error, message in self.items.items():
            messages += message

        return messages

    def reset(self):
        """Gets a list of all the messages"""
        self.items = {}

    def merge(self, dictionary):
        """Merge a dictionary into the message bag.

        Arguments:
            dictionary {dict} -- dictionary of errors and messages.

        Returns:
            dictionary -- Returns a dictionary of the new errors.
        """
        self.items.update(dictionary)
        return self.any()

    def new(self, dictionary):
        return self.__class__(dictionary)

    def __len__(self):
        return len(self.items)

    def __str__(self):
        return json.dumps(self.items)

    def get_response(self):
        return json.dumps(self.items)

    @staticmethod
    def view_helper(errors={}):
        if errors:
            return MessageBag(errors)

        from wsgi import application

        return MessageBag(application.make("request").session.get("errors") or {})



from src.masonite.environment import env

KEY = env("APP_KEY", "-RkDOqXojJIlsF_I8wWiUq_KRZ0PtGWTOZ676u5HtLg=")

DEBUG = env("APP_DEBUG", True)

HASHING = {
    "default": "bcrypt",
    "bcrypt": {"rounds": 10},
    "argon2": {"memory": 1024, "threads": 2, "time": 2},
}

APP_URL = env("APP_URL", "http://localhost:8000/")

MIX_BASE_URL = env("MIX_BASE_URL", None)



"""Notification handler class"""
import uuid

from ..exceptions.exceptions import NotificationException
from .AnonymousNotifiable import AnonymousNotifiable


class NotificationManager:
    """Notification handler which handle sending/queuing notifications anonymously
    or to notifiables through different channels."""

    sent_notifications = {}
    dry_notifications = {}

    def __init__(self, application, driver_config=None):
        self.application = application
        self.drivers = {}
        self.driver_config = driver_config or {}
        self.options = {"dry": False}

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})
        self.get_driver(name).set_options(self.get_config_options(name))

    def get_driver(self, name):
        return self.drivers[name]

    def set_configuration(self, config):
        self.driver_config = config.get("drivers")
        self.options.update({"dry": config.get("dry")})
        return self

    def get_config_options(self, driver):
        return self.driver_config.get(driver, {})

    def send(
        self, notifiables, notification, drivers=[], dry=False, fail_silently=False
    ):
        """Send the given notification to the given notifiables."""
        notifiables = self._format_notifiables(notifiables)
        if not notification.should_send() or dry or self.options.get("dry"):
            key = notification.type()
            self.dry_notifications.update(
                {key: notifiables + self.dry_notifications.get(key, [])}
            )
            return
        results = []
        for notifiable in notifiables:
            # get drivers to use for sending this notification
            drivers = drivers if drivers else notification.via(notifiable)
            if not drivers:
                raise NotificationException(
                    "No drivers have been defined in via() method of {0} notification.".format(
                        notification.type()
                    )
                )
            notification.id = uuid.uuid4()
            for driver in drivers:
                driver_instance = self.get_driver(driver)
                if isinstance(notifiable, AnonymousNotifiable) and driver == "database":
                    # this case is not possible but that should not stop other channels to be used
                    continue
                try:
                    # if isinstance(notification, ShouldQueue):
                    #     results.append(driver_instance.queue(notifiable, notification))
                    # else:
                    results.append(driver_instance.send(notifiable, notification))
                except Exception as e:
                    if not notification.ignore_errors() and not fail_silently:
                        raise e

        return results[0] if len(results) == 1 else results

    def _format_notifiables(self, notifiables):
        from masoniteorm.collection import Collection

        if isinstance(notifiables, (list, tuple, Collection)):
            return notifiables
        else:
            return [notifiables]

    def route(self, driver, route):
        """Specify how to send a notification to an anonymous notifiable."""
        return AnonymousNotifiable(self.application).route(driver, route)



from src.masonite.controllers import Controller
from src.masonite.views import View
from src.masonite.mail import Mailable, Mail


class Welcome(Mailable):
    def build(self):
        return (
            self.to("idmann509@gmail.com")
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
            .html("<h1>Hello from Masonite!</h1>")
        )


class MailableController(Controller):
    def view(self, mail: Mail):
        mail.mailable(Welcome()).send(driver="mailgun")



import inspect


class TaskHandler:
    def __init__(self, application, tasks=None):
        if tasks is None:
            tasks = []

        self.tasks = tasks
        self.application = application

    def add(self, *tasks):
        self.tasks += list(tasks)

    def run(self, run_name=None, force=False):
        app = self.application
        for task_class in self.tasks:
            # Resolve the task with the container
            task_identifier = [task_class.name, task_class.__class__.__name__]
            if run_name and run_name not in task_identifier:
                continue

            if inspect.isclass(task_class):
                task = app.resolve(task_class)
            else:
                task = task_class

            # If the class should run then run it
            if task.should_run() or force:
                task.handle()



import re
import tldextract
from typing import TYPE_CHECKING, List, Any

from ..cookies import CookieJar
from ..headers import HeaderBag, Header
from ..input import InputBag
from .validation import ValidatesRequest
from ..authorization import AuthorizesRequest
from ..sessions import old as old_helper

if TYPE_CHECKING:
    from ..routes import Route


class Request(ValidatesRequest, AuthorizesRequest):
    def __init__(self, environ: dict):
        """Initializes a request with the WSGI environment dictionary."""
        self.environ = environ
        self.cookie_jar = CookieJar()
        self.header_bag = HeaderBag()
        self.input_bag = InputBag()
        self._ip = None
        self.params = {}
        self._user = None
        self.route = None
        self._subdomains_activated = False
        self.load()

    def load(self):
        """Load request from environment."""
        self.cookie_jar.load(self.environ.get("HTTP_COOKIE", ""))
        self.header_bag.load(self.environ)
        self.input_bag.load(self.environ)

    def load_params(self, params: dict = None):
        """Load request parameters."""
        if not params:
            params = {}

        self.params.update(params)

    def param(self, param: str, default: str = "") -> str:
        """Get query string parameter from request."""
        return self.params.get(param, default)

    def get_route(self) -> "Route":
        """Get Route associated to request if any."""
        return self.route

    def get_path(self) -> str:
        """Get request path (read from PATH_INFO) environment variable without eventual query
        string parameters."""
        return self.environ.get("PATH_INFO")

    def get_path_with_query(self) -> str:
        """Get request path (read from PATH_INFO) environment variable with eventual query
        string parameters."""
        query_string = self.environ.get("QUERY_STRING")
        if query_string:
            return self.get_path() + "?" + query_string
        else:
            return self.get_path()

    def get_back_path(self) -> str:
        """Get previous request path if it has been defined as '__back' input."""
        return self.input("__back") or self.get_path_with_query()

    def get_request_method(self) -> str:
        """Get request method (read from REQUEST_METHOD environment variable)."""
        return self.input("__method") or self.environ.get("REQUEST_METHOD")

    def input(self, name: str, default: str = "") -> str:
        """Get a specific request input value with the given name. If the value does not exist in
        the request return the default value."""
        name = str(name)

        return self.input_bag.get(name, default=default)

    def cookie(self, name: str, value: str = None, **options) -> None:
        """If no value provided, read the cookie value with the given name from the request. Else
        create a cookie in the request with the given name and value.
        Some options can be passed when creating cookie, refer to CookieJar class."""
        if value is None:
            cookie = self.cookie_jar.get(name)
            if not cookie:
                return
            return cookie.value

        return self.cookie_jar.add(name, value, **options)

    def delete_cookie(self, name: str) -> "Request":
        """Delete cookie with the given name from the request."""
        self.cookie_jar.delete(name)
        return self

    def header(self, name: str, value: str = None) -> "str|None":
        """If no value provided, read the header value with the given name from the request. Else
        add a header in the request with the given name and value."""
        if value is None:
            header = self.header_bag.get(name)
            if not header:
                return
            return header.value
        else:
            return self.header_bag.add(Header(name, value))

    def all(self) -> dict:
        """Get all inputs from the request as a dictionary."""
        return self.input_bag.all_as_values()

    def only(self, *inputs: List[str]) -> dict:
        """Get only the given inputs from the request as a dictionary."""
        return self.input_bag.only(*inputs)

    def old(self, key: str):
        """Get value from session for the given key."""
        return old_helper(key)

    def is_not_safe(self) -> bool:
        """Check if the current request is considered 'safe', meaning that the request method is
        GET, OPTIONS or HEAD."""
        if not self.get_request_method() in ("GET", "OPTIONS", "HEAD"):
            return True

        return False

    def user(self) -> "None|Any":
        """Get the current authenticated user if any. LoadUserMiddleware needs to be used for user
        to be populated in request."""
        return self._user

    def set_user(self, user: Any) -> "Request":
        """Set the current authenticated user of the request."""
        self._user = user
        return self

    def remove_user(self) -> "Request":
        """Log out user of the current request."""
        self._user = None
        return self

    def contains(self, route: str) -> bool:
        """Check if current request path match the given URL."""
        if not route.startswith("/"):
            route = "/" + route

        regex = re.compile(route.replace("*", "[a-zA-Z0-9_]+"))

        return regex.match(self.get_path())

    def get_subdomain(self, exclude_www: bool = True) -> "None|str":
        """Get the request subdomain if subdomains are enabled."""
        if not self._subdomains_activated:
            return None

        url = tldextract.extract(self.get_host())
        if url.subdomain == "" or (
            url.subdomain and exclude_www and url.subdomain == "www"
        ):
            return None

        return url.subdomain

    def get_host(self) -> str:
        """Get the request host (from HTTP_HOST environment variable)."""
        return self.environ.get("HTTP_HOST")

    def activate_subdomains(self):
        """Enable subdomains for this request."""
        self._subdomains_activated = True
        return self

    def ip(self) -> "str|None":
        """Return the request IP by processing the different headers setup in IpMiddleware."""
        return self._ip

    def accepts_json(self) -> bool:
        """Check if request Accept header contains application/json."""
        return "application/json" in str(self.header("Accept"))



"""New Preset Command."""
from .Command import Command
from ..presets import Remove


class PresetCommand(Command):
    """
    Scaffold frontend preset in your project

    preset
        {name? : Name of the preset}
        {--r|remove=? : Remove all scaffolded presets}
        {--l|list=? : List all available presets}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        preset_name = self.argument("name")
        remove_presets = self.option("remove")
        list_presets = self.option("list")

        presets = self.app.make("presets")
        available_presets = presets.get_presets_keys()

        if list_presets:
            return self.info(f"Available presets are: {', '.join(available_presets)}")

        if remove_presets:
            return self.remove()

        if preset_name not in available_presets:
            return self.error(
                f"Invalid preset. Available presets are: {', '.join(available_presets)}"
            )

        self.info(f"Scaffolding {preset_name} preset...")
        presets.run(preset_name)
        self.info(f"{preset_name} installed successfully.")
        self.comment(
            'Please run "npm install && npm run dev" to compile your fresh scaffolding.'
        )

    def remove(self):
        Remove().install()
        self.info("Frontend scaffolding removed successfully.")



class TestResponseCapsule:
    def __init__(self, base_test_response_class):
        self.base_test_response_class = base_test_response_class
        self.test_responses_classes = []

    def add(self, *classes):
        self.test_responses_classes.extend(classes)
        return self

    def build(self, *args):
        """Apply other test response class as mixins to base test response class."""
        obj = self.base_test_response_class(*args)
        for cls in self.test_responses_classes:
            base_cls = obj.__class__
            base_cls_name = obj.__class__.__name__
            obj.__class__ = type(base_cls_name, (base_cls, cls), {})
        return obj



from typing import TYPE_CHECKING, List, Any

if TYPE_CHECKING:
    from ..routes import Route

class Request:
    """Request facade."""

    def load(self):
        """Load request from environment."""
        ...
    def load_params(params: dict = None):
        """Load request parameters."""
        ...
    def param(param: str, default: str = "") -> str:
        """Get query string parameter from request."""
        ...
    def get_route(self) -> "Route":
        """Get Route associated to request if any."""
        ...
    def get_path(self) -> str:
        """Get request path (read from PATH_INFO) environment variable without eventual query
        string parameters."""
        ...
    def get_path_with_query(self) -> str:
        """Get request path (read from PATH_INFO) environment variable with eventual query
        string parameters."""
        ...
    def get_back_path(self) -> str:
        """Get previous request path if it has been defined as '__back' input."""
        ...
    def get_request_method(self) -> str:
        """Get request method (read from REQUEST_METHOD environment variable)."""
        ...
    def input(name: str, default: str = "") -> str:
        """Get a specific request input value with the given name. If the value does not exist in
        the request return the default value."""
        ...
    def cookie(name: str, value: str = None, **options) -> None:
        """If no value provided, read the cookie value with the given name from the request. Else
        create a cookie in the request with the given name and value.
        Some options can be passed when creating cookie, refer to CookieJar class."""
        ...
    def delete_cookie(name: str) -> "Request":
        """Delete cookie with the given name from the request."""
        ...
    def header(name: str, value: str = None) -> "str|None":
        """If no value provided, read the header value with the given name from the request. Else
        add a header in the request with the given name and value."""
        ...
    def all(self) -> dict:
        """Get all inputs from the request as a dictionary."""
        ...
    def only(*inputs: List[str]) -> dict:
        """Get only the given inputs from the request as a dictionary."""
        ...
    def old(key: str):
        """Get value from session for the given key."""
        ...
    def is_not_safe(self) -> bool:
        """Check if the current request is considered 'safe', meaning that the request method is
        GET, OPTIONS or HEAD."""
        ...
    def user(self) -> "None|Any":
        """Get the current authenticated user if any. LoadUserMiddleware needs to be used for user
        to be populated in request."""
        ...
    def set_user(user: Any) -> "Request":
        """Set the current authenticated user of the request."""
        ...
    def remove_user(self) -> "Request":
        """Log out user of the current request."""
        ...
    def contains(route: str) -> bool:
        """Check if current request path match the given URL."""
        ...
    def get_subdomain(exclude_www: bool = True) -> "None|str":
        """Get the request subdomain if subdomains are enabled."""
        ...
    def get_host(self) -> str:
        """Get the request host (from HTTP_HOST environment variable)."""
        ...
    def activate_subdomains(self):
        """Enable subdomains for this request."""
        ...



from .. import Middleware
from ...facades import Session
from ...validation import MessageBag


class ShareErrorsInSessionMiddleware(Middleware):
    """Share errors as a Message bag if there are any errors flashed to session. If not the message
    bag will be empty."""

    def before(self, request, _):
        request.app.make("view").share(
            {
                "errors": MessageBag(Session.pull("errors") or {}),
            }
        )
        return request

    def after(self, request, _):
        return request



from tests import TestCase
from src.masonite.facades import Hash


class TestHashers(TestCase):
    def test_should_hash_and_check_using_hasher(self):

        for driver in Hash.drivers.keys():
            hashed = Hash.make("masonite", driver=driver)
            assert hashed != "masonite"
            assert Hash.check("masonite", hashed, driver=driver)

    def test_bcrypt_needs_rehash(self):
        hashed = Hash.make("masonite", options={"rounds": 5})
        # here no options is given so default rounds will be used (10 in tests config)
        assert Hash.needs_rehash(hashed)

    def test_argon2_needs_rehash(self):
        hashed = Hash.make(
            "masonite",
            driver="argon2",
            options={
                "memory": 512,
                "threads": 8,
                "time": 2,
            },
        )
        # Here argon2 method is invoked without custom options and will
        # use default argon2 configuration, so rehash will be needed
        assert Hash.needs_rehash(hashed, driver="argon2")

    def test_should_return_hash_content_as_string(self):

        for driver in Hash.drivers.keys():
            hashed = Hash.make("masonite", driver=driver)
            assert type(hashed) == str

    def test_should_return_hash_content_as_bytes(self):

        for driver in Hash.drivers.keys():
            hashed = Hash.make_bytes("masonite", driver=driver)
            assert type(hashed) == bytes



DRIVERS = {
    "default": "async",
    "database": {
        "connection": "sqlite",
        "table": "jobs",
        "failed_table": "failed_jobs",
        "attempts": 3,
        "poll": 5,
        "tz": "UTC",
    },
    "redis": {
        #
    },
    "amqp": {
        "username": "guest",
        "password": "guest",
        "port": "5672",
        "vhost": "",
        "host": "localhost",
        # See https://pika.readthedocs.io/en/stable/modules/parameters.html#pika.connection.URLParameters
        # for valid connection options values
        "connection_options": {},
        "exchange": "",
        "channel": "default",
        "queue": "masonite4",
        "tz": "UTC",
    },
    "async": {
        "blocking": True,
        "callback": "handle",
        "mode": "threading",
        "workers": 1,
    },
}



from inspect import isclass

from ..response import Response
from ..facades import Response as ResponseFacade
from .Provider import Provider
from ..routes import Route
from ..routes.commands import RouteListCommand
from ..pipeline import Pipeline
from ..exceptions import RouteNotFoundException


class RouteProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        Route.set_controller_locations(self.application.make("controllers.location"))
        self.application.make("commands").add(RouteListCommand(self.application))

    def boot(self):
        router = self.application.make("router")
        request = self.application.make("request")
        response = self.application.make("response")

        route = router.find(
            request.get_path(), request.get_request_method(), request.get_subdomain()
        )

        request.route = route

        # Run before middleware
        before_middleware = Pipeline(request, response).through(
            self.application.make("middleware").get_http_middleware(),
            handler="before",
        )

        exception = None
        if before_middleware:
            if route:
                request.load_params(route.extract_parameters(request.get_path()))
                route_middleware = self.application.make(
                    "middleware"
                ).run_route_middleware(
                    route.get_middlewares(), request, response, callback="before"
                )

                if route_middleware:
                    try:
                        data = route.get_response(self.application)
                        if isinstance(data, Response) or (
                            isclass(data) and issubclass(data, ResponseFacade)
                        ):
                            pass
                        else:
                            response.view(data)
                    except Exception as e:
                        exception = e

                    self.application.make("middleware").run_route_middleware(
                        route.get_middlewares(), request, response, callback="after"
                    )

            else:
                raise RouteNotFoundException(
                    f"{request.get_request_method()} {request.get_path()} : 404 Not Found"
                )

        Pipeline(request, response).through(
            self.application.make("middleware").get_http_middleware(),
            handler="after",
        )

        if exception:
            raise exception



import pytest
import pendulum
from src.masonite.scheduling import TaskHandler, Task
from tests import TestCase


class MockTask1(Task):
    run_every = "1 minutes"
    timezone = "America/New_York"

    def handle(self):
        print("hello 1")


class MockTask2(Task):
    run_every = "1 minutes"
    timezone = "America/New_York"

    def handle(self):
        print("hello 2")


class TestHandler(TestCase):
    def test_handler_adds_and_runs_tasks(self):
        self.handler = TaskHandler(self.application)
        self.handler.add(MockTask1, MockTask2())
        self.handler.run()



from inspect import isclass
import re
from urllib import parse

from ..utils.str import modularize, removeprefix
from ..exceptions import InvalidRouteCompileException
from ..facades import Loader
from ..controllers import Controller
from ..exceptions import LoaderNotFound


class HTTPRoute:
    def __init__(
        self,
        url,
        controller=None,
        request_method=["get"],
        name=None,
        compilers=None,
        controllers_locations=["app.http.controllers"],
        controller_bindings=[],
        **options,
    ):
        if not url.startswith("/"):
            url = "/" + url
        if url.endswith("/") and url != "/":
            url = url[:-1]

        self.url = url
        self.controllers_locations = controllers_locations
        self.controller = controller
        self.controller_class = None
        self.controller_instance = None
        self.controller_method = None
        self._domain = None
        self._name = name
        self.request_method = [x.lower() for x in request_method]
        self.list_middleware = []
        self.excluded_middlewares = []
        self.e = None
        self.compilers = compilers or {}
        self._find_controller(controller)
        self.controller_bindings = controller_bindings
        self.compile_route_to_regex()

    def __str__(self):
        return f"<HttpRoute [{self._name}]: {self.url}>"

    def match(self, path, request_method, subdomain=None):

        route_math = (
            re.match(self._compiled_regex, path)
            or re.match(self._compiled_regex_end, path)
        ) and request_method.lower() in self.request_method

        domain_match = subdomain == self._domain

        return route_math and domain_match

    def get_name(self):
        return self._name

    def matches(self, path):
        return re.match(self._compiled_regex, path) or re.match(
            self._compiled_regex_end, path
        )

    def match_name(self, name):
        return name == self._name

    def name(self, name):
        self._name = name
        return self

    def domain(self, subdomain):
        self._domain = subdomain
        return self

    def to_url(self, parameters: dict = {}, query_params: dict = {}) -> str:
        """Transform route to a URL string with the given url and query parameters."""

        # Split the url into a list
        split_url = self.url.split("/")

        # Start beginning of the new compiled url
        compiled_url = "/"

        # Iterate over the list
        for url in split_url:
            if url:
                # if the url contains a parameter variable like @id:int
                if "@" in url:
                    url = url.replace("@", "").split(":")[0]
                    if isinstance(parameters, dict):
                        compiled_url += str(parameters[url]) + "/"
                    elif isinstance(parameters, list):
                        compiled_url += str(parameters.pop(0)) + "/"
                elif "?" in url:
                    url = url.replace("?", "").split(":")[0]
                    if isinstance(parameters, dict):
                        compiled_url += str(parameters.get(url, "/")) + "/"
                    elif isinstance(parameters, list):
                        compiled_url += str(parameters.pop(0)) + "/"
                else:
                    compiled_url += url + "/"

        # The loop isn't perfect and may have an unwanted trailing slash
        if compiled_url.endswith("/"):
            compiled_url = compiled_url[:-1]

        # The loop isn't perfect and may have 2 slashes next to eachother
        if "//" in compiled_url:
            compiled_url = compiled_url.replace("//", "/")

        # Add eventual query parameters
        if query_params:
            compiled_url += "?" + parse.urlencode(query_params)
        return compiled_url

    def _find_controller(self, controller):
        """Find the controller to attach to the route. Look for controller (str or class) in all
        specified controllers_location.

        Arguments:
            controller {string|object} -- String or object controller to search for.

        Returns:
            None
        """
        if controller is None:
            return None
        # If the output specified is a string controller e.g. "WelcomeController@show"
        elif isinstance(controller, str):
            if "@" in controller:
                controller_path, controller_method_str = controller.split("@")
            else:
                controller_path = controller
                controller_method_str = "__call__"

            controller_path = modularize(controller_path).split(".")
            if len(controller_path) > 1:
                controller_name = controller_path.pop()
                prefix_path = ".".join(controller_path)
            else:
                controller_name = controller_path[0]
                prefix_path = ""
            # build a list of all locations where the controller can be found
            # if the controller is defined such as auth.WelcomeController, append the prefix path to
            # the locations
            locations = list(
                map(
                    lambda loc: f"{loc}.{removeprefix(prefix_path, loc)}"
                    if prefix_path
                    else loc,
                    self.controllers_locations,
                )
            )
            try:
                self.controller_class = Loader.find(
                    Controller, locations, controller_name, raise_exception=True
                )
            except LoaderNotFound as e:
                self.e = e
                print("\033[93mTrouble importing controller!", str(e), "\033[0m")
        # controller is an instance with a bound method
        elif hasattr(controller, "__self__"):
            _, controller_method_str = controller.__qualname__.split(".")
            self.controller_instance = controller.__self__

        # it's a class or class.method, we don't have to find it, just get the class
        elif hasattr(controller, "__qualname__"):
            if "." in controller.__qualname__:
                controller_name, controller_method_str = controller.__qualname__.split(
                    "."
                )
            else:
                controller_name = controller.__qualname__
                controller_method_str = "__call__"

            try:
                self.controller_class = Loader.get_object(
                    controller.__module__, controller_name, raise_exception=True
                )
            except LoaderNotFound as e:
                self.e = e
                print("\033[93mTrouble importing controller!", str(e), "\033[0m")
        # it's a controller instance
        else:
            self.controller_instance = controller
            controller_method_str = "__call__"

        # Set the controller method on class. This is a string
        self.controller_method = controller_method_str

    def get_response(self, app=None):
        # Resolve Controller Constructor
        if self.e:
            print(
                "\033[93mCannot find controller {}. Did you create this one?".format(
                    self.controller
                ),
                "\033[0m",
            )
            raise SyntaxError(str(self.e))

        if app:
            if self.controller_instance:
                controller = self.controller_instance
            else:
                controller = app.resolve(
                    self.controller_class, *self.controller_bindings
                )
            # resolve route parameters
            params = self.extract_parameters(app.make("request").get_path()).values()
            # Resolve Controller Method
            response = app.resolve(getattr(controller, self.controller_method), *params)
            return response

        if self.controller_instance:
            controller = self.controller_instance
        else:
            controller = self.controller_class()

        return getattr(controller, self.controller_method)()

    def middleware(self, *args):
        """Load a list of middleware to run.

        Returns:
            self
        """
        for arg in args:
            if arg and arg not in self.list_middleware:
                self.list_middleware.append(arg)
        return self

    def prepend_middleware(self, *args):
        """Load a list of middleware to run.

        Returns:
            self
        """
        for arg in args:
            if arg and arg not in self.list_middleware:
                self.list_middleware.insert(0, arg)
        return self

    def get_middlewares(self):
        """Get all the middlewares to run for this route."""
        middlewares = self.list_middleware

        for middleware in self.excluded_middlewares:
            if middleware in middlewares:
                middlewares.remove(middleware)

        return middlewares

    def set_middleware(self, middleware):
        """Get all the middlewares to run for this route."""
        self.list_middleware = middleware

        return self

    def exclude_middleware(self, *args):
        """Remove a list of middleware for this route. It can be useful when
        using Route group middleware to override middleware for a given route in the group."""
        for middleware in args:
            self.excluded_middlewares.append(middleware)
        return self

    def compile_route_to_regex(self):
        """Compile the given route to a regex string.

        Arguments:
            route {string} -- URI of the route to compile.

        Returns:
            string -- Compiled URI string.
        """
        # Split the route
        split_given_route = self.url.split("/")
        # compile the provided url into regex
        url_list = []
        regex = "^"
        for regex_route in split_given_route:
            if "@" in regex_route:
                if ":" in regex_route:
                    try:
                        regex += self.compilers[regex_route.split(":")[1]]
                    except KeyError:
                        raise InvalidRouteCompileException(
                            'Route compiler "{}" is not an available route compiler. '
                            "Verify you spelled it correctly or that you have added it using the compile() method.".format(
                                regex_route.split(":")[1]
                            )
                        )

                else:
                    regex += self.compilers["default"]

                regex += r"\/"

                # append the variable name passed @(variable):int to a list
                url_list.append(regex_route.replace("@", "").split(":")[0])
            elif "?" in regex_route:
                # Make the preceding token match 0 or more
                regex += "?"

                if ":" in regex_route:

                    try:
                        regex += self.compilers[regex_route.split(":")[1]] + "*"
                    except KeyError:
                        if self.request:
                            raise InvalidRouteCompileException(
                                'Route compiler "{}" is not an available route compiler. '
                                "Verify you spelled it correctly or that you have added it using the compile() method.".format(
                                    regex_route.split(":")[1]
                                )
                            )
                        self._compiled_regex = None
                        self._compiled_regex_end = None
                        return

                else:
                    regex += self.compilers["default"] + "*"

                regex += r"\/"

                url_list.append(regex_route.replace("?", "").split(":")[0])
            else:
                regex += regex_route + r"\/"

        self.url_list = url_list
        regex += "$"
        self._compiled_regex = re.compile(regex.replace(r"\/$", r"$"))
        self._compiled_regex_end = re.compile(regex)

        return regex

    def extract_parameters(self, path):
        if not self.url_list:
            return {}

        if (not path.endswith("/")) or path == "/":
            matching_regex = self._compiled_regex
        else:
            matching_regex = self._compiled_regex_end
        return dict(zip(self.url_list, matching_regex.match(path).groups()))



from .CommandCapsule import CommandCapsule
from .AuthCommand import AuthCommand
from .TinkerCommand import TinkerCommand
from .KeyCommand import KeyCommand
from .ServeCommand import ServeCommand
from .QueueWorkCommand import QueueWorkCommand
from .QueueRetryCommand import QueueRetryCommand
from .QueueTableCommand import QueueTableCommand
from .QueueFailedCommand import QueueFailedCommand
from .MakeControllerCommand import MakeControllerCommand
from .MakeJobCommand import MakeJobCommand
from .MakeMailableCommand import MakeMailableCommand
from .MakeProviderCommand import MakeProviderCommand
from .PublishPackageCommand import PublishPackageCommand
from .MakePolicyCommand import MakePolicyCommand
from .MakeTestCommand import MakeTestCommand
from .DownCommand import DownCommand
from .UpCommand import UpCommand
from .MakeCommandCommand import MakeCommandCommand
from .MakeViewCommand import MakeViewCommand
from .MakeMiddlewareCommand import MakeMiddlewareCommand
from .PresetCommand import PresetCommand
from .Command import Command



from src.masonite.notification import Notification, Textable
from src.masonite.mail import Mailable
from src.masonite.mail import Mailable


class OneTimePassword(Notification, Mailable, Textable):
    def to_mail(self, notifiable):
        return (
            self.to(notifiable.email)
            .subject("Masonite 4")
            .from_("hello@email.com")
            .text(f"Hello {notifiable.name}")
        )

    def to_vonage(self, notifiable):
        return self.text_message("Welcome !").to("6314870798").from_("33123456789")

    def via(self, notifiable):
        return ["vonage"]



from .FrameworkProvider import FrameworkProvider
from .RouteProvider import RouteProvider
from .ViewProvider import ViewProvider
from .WhitenoiseProvider import WhitenoiseProvider
from .ExceptionProvider import ExceptionProvider
from .AuthenticationProvider import AuthenticationProvider
from .AuthorizationProvider import AuthorizationProvider
from .Provider import Provider
from .MailProvider import MailProvider
from .SessionProvider import SessionProvider
from .HelpersProvider import HelpersProvider
from .QueueProvider import QueueProvider
from .CacheProvider import CacheProvider
from ..events.providers import EventProvider
from ..filesystem.providers import StorageProvider
from ..broadcasting.providers import BroadcastProvider
from ..scheduling.providers import ScheduleProvider
from ..essentials.providers.HashIDProvider import HashIDProvider
from .HashServiceProvider import HashServiceProvider
from ..validation.providers import ValidationProvider
from ..configuration.providers import ConfigurationProvider
from ..orm.providers import ORMProvider
from ..presets.providers import PresetsProvider



from tests import TestCase

from src.masonite.utils.structures import data_get, data_set, data


class TestStructures(TestCase):
    def test_data_get(self):
        struct = {"key": "val", "a": {"b": "c", "nested": {"a": 1}}}
        self.assertEqual(data_get(struct, "key"), "val")

        self.assertEqual(data_get(struct, "a.b"), "c")
        self.assertEqual(data_get(struct, "a.nested.a"), 1)

        self.assertEqual(data_get(struct, "a.nested.unknown"), None)
        self.assertEqual(data_get(struct, "a.nested.unknown", 0), 0)

    def test_data_set(self):
        struct = {"key": "val", "a": {"b": "c", "nested": {"a": 1}}}
        data_set(struct, "key", "val2")
        self.assertEqual(struct.get("key"), "val2")

        data_set(struct, "a.nested.a", 3)
        self.assertEqual(data_get(struct, "a.nested.a"), 3)

        data_set(struct, "a.unknown", "new")
        self.assertEqual(data_get(struct, "a.unknown"), "new")

    def test_data_set_no_overwrite(self):
        struct = {"key": "val", "a": {"b": "c", "nested": {"a": 1}}}
        data_set(struct, "key", "val2", overwrite=False)
        self.assertEqual(data_get(struct, "key"), "val")

        data_set(struct, "a.nested.a", "new", overwrite=False)
        self.assertEqual(data_get(struct, "a.nested.a"), 1)

        data_set(struct, "unknown.key", "new", overwrite=False)
        self.assertEqual(data_get(struct, "unknown.key"), "new")

    def test_data(self):
        struct = {"key": "val"}
        dotted_struct = data(struct)
        dotted_struct["new_key.nested"] = 3
        self.assertEqual(dotted_struct.get("new_key.nested"), 3)
        self.assertEqual(dotted_struct, {"key": "val", "new_key": {"nested": 3}})



from src.masonite.controllers import Controller
from src.masonite.views import View
from src.masonite.request import Request
from src.masonite.response import Response
from src.masonite.authentication import Auth


class LoginController(Controller):
    def show(self, view: View):
        return view.render("auth.login")

    def store(self, view: View, request: Request, auth: Auth, response: Response):
        login = auth.attempt(request.input("username"), request.input("password"))

        if login:
            return response.redirect(name="home")

        # Go back to login page
        return response.redirect(name="login")



"""Craft Command.

This module is really used for backup only if the masonite CLI cannot import this for you.
This can be used by running "python craft". This module is not ran when the CLI can
successfully import commands for you.
"""

from cleo import Application

from .. import __version__
from .ProjectCommand import (
    ProjectCommand,
)
from .KeyCommand import KeyCommand
from .InstallCommand import InstallCommand

application = Application("Masonite Starter", __version__)

application.add(ProjectCommand())
application.add(KeyCommand())
application.add(InstallCommand())

if __name__ == "__main__":
    application.run()



from typing import Any, List, TYPE_CHECKING

if TYPE_CHECKING:
    from ..dumps import Dump as DumpObject

class Dump:
    """Dumper facade."""

    def clear(self) -> "Dump":
        """Clear all dumped data"""
        ...
    def dd(*objects: List[Any]):
        """Dump all provided args and die, raising a DumpException."""
        ...
    def dump(*objects: List[Any]):
        """Dump all provided args and continue code execution. This does not raise a DumpException."""
        ...
    def get_dumps(ascending: bool = False) -> List[DumpObject]:
        """Get all dumps as Dump objects. If ascending is True, get dumps from oldest to most recents."""
        ...
    def last(self) -> DumpObject:
        """Return last added dump."""
        ...
    def get_serialized_dumps(ascending: bool = False) -> List[dict]:
        """Get all dumps as dict. If ascending is True, sort dumps from oldest to most recents."""
        ...



from tests import TestCase
import os
import time
from src.masonite.filesystem import File
import pytest


@pytest.mark.integrations
class TestLocalStorage(TestCase):
    def setUp(self):
        super().setUp()
        self.application.make("storage")
        self.driver = self.application.make("storage").disk("s3")

    def test_can_get_file_driver(self):
        self.driver.put("key.log", "value")
        self.assertEqual(self.driver.get("key.log"), "value")
        self.assertEqual(self.driver.get("key.logs"), None)
        self.assertTrue(self.driver.exists("key.log"))

    def test_can_move(self):
        self.driver.move("key.log", "logs/key.log")

    def test_can_stream(self):
        self.driver.put("key.log", "value")
        stream = self.driver.stream("key.log")
        self.assertEqual(stream.name(), "key.log")
        self.assertEqual(stream.extension(), ".log")

    def test_can_delete(self):
        self.driver.put("delete.log", "value")
        self.assertEqual(self.driver.get("delete.log"), "value")
        stream = self.driver.delete("delete.log")
        self.assertEqual(self.driver.get("delete.log"), None)

    def test_can_store(self):
        self.driver.store(File(b"hello.log", "hello-world"))

    def test_can_append_file(self):
        self.driver.append("world.log", "hello")
        self.driver.prepend("world.log", "world")

    def test_can_get_contents_of_directory(self):
        self.driver.get_files()



from tests import TestCase

from src.masonite.notification import Notification, Notifiable
from src.masonite.mail import Mailable
from masoniteorm.models import Model


class User(Model, Notifiable):
    """User Model"""

    __fillable__ = ["name", "email", "password"]


class WelcomeNotification(Notification):
    def to_mail(self, notifiable):
        return (
            Mailable()
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
        )

    def via(self, notifiable):
        return ["mail"]


class TestNotification(TestCase):
    def test_should_send(self):
        notification = WelcomeNotification()
        self.assertTrue(notification.should_send())

    def test_ignore_errors(self):
        notification = WelcomeNotification()
        self.assertFalse(notification.ignore_errors())

    def test_notification_type(self):
        self.assertEqual("WelcomeNotification", WelcomeNotification().type())


DRY = True


class TestNotificationManager(TestCase):
    def test_dry_mode(self):
        # locally when sending to anonymous or notifiable
        self.assertEqual(
            self.application.make("notification")
            .route("mail", "test@mail.com")
            .send(WelcomeNotification(), dry=True),
            None,
        )

        user = User.find(1)
        user.notify(WelcomeNotification(), dry=True)

        # globally
        # override settings for testing purposes
        self.assertEqual(
            self.application.make("notification")
            .route("mail", "test@mail.com")
            .send(WelcomeNotification(), dry=True),
            None,
        )



from tests import TestCase

from src.masonite.loader.Loader import Loader
from src.masonite.exceptions import LoaderNotFound

OBJ_1 = "test"


class TestLoader(TestCase):
    def setUp(self):
        super().setUp()
        self.loader = Loader()

    def test_get_objects(self):
        objects = self.loader.get_objects("tests.features.loader.test_loader")
        self.assertIsInstance(objects, dict)
        self.assertEqual(objects.get("OBJ_1"), "test")
        self.assertEqual(objects.get("Loader"), Loader)

    def test_get_objects_for_unexisting_path(self):
        objects = self.loader.get_objects("test.not.existing.path")
        self.assertIsNone(objects)

    def test_get_parameters(self):
        objects = self.loader.get_parameters("tests.features.loader.test_loader")
        self.assertIsInstance(objects, dict)
        self.assertEqual(objects.get("OBJ_1"), "test")
        self.assertEqual(len(objects.keys()), 1)

    def test_find_all(self):
        from masoniteorm.models import Model
        from tests.integrations.app.User import User

        objects = self.loader.find_all(Model, "tests.integrations.app")
        self.assertIsInstance(objects, dict)
        self.assertEqual(objects.get("User"), User)
        self.assertEqual(len(objects.keys()), 1)

    def test_find(self):
        from masoniteorm.models import Model
        from tests.integrations.app.User import User

        obj = self.loader.find(Model, "tests.integrations.app", "User")
        self.assertEqual(obj, User)

    def test_find_methods_raise_exception_if_specified(self):
        from masoniteorm.models import Model

        with self.assertRaises(LoaderNotFound):
            self.loader.find(
                Model, "test.not.existing.path", "User", raise_exception=True
            )

        with self.assertRaises(LoaderNotFound):
            self.loader.find_all(Model, "test.not.existing.path", raise_exception=True)



from ..authentication import Auth
from ..authentication.guards import WebGuard
from ..configuration import config
from .Provider import Provider


class AuthenticationProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        auth = Auth(self.application).set_configuration(config("auth.guards"))
        auth.add_guard("web", WebGuard(self.application))
        self.application.bind("auth", auth)

    def boot(self):
        pass



from .MessageAttachment import MessageAttachment


class Mailable:
    def __init__(self):
        self._to = ""
        self._cc = ""
        self._bcc = ""
        self._from = ""
        self._reply_to = ""
        self._subject = ""
        self._priority = None
        self._driver = None
        self.text_content = ""
        self.html_content = ""
        self.attachments = []
        self.headers = {}

    def to(self, to):
        self._to = to
        return self

    def cc(self, cc):
        self._cc = cc
        return self

    def bcc(self, bcc):
        self._bcc = bcc
        return self

    def header(self, key, value):
        self.headers.update({key: value})
        return self

    def set_application(self, application):
        self.application = application
        return self

    def from_(self, _from):
        self._from = _from
        return self

    def attach(self, name, path):
        self.attachments.append(MessageAttachment(name, path))
        return self

    def reply_to(self, reply_to):
        self._reply_to = reply_to
        return self

    def subject(self, subject):
        self._subject = subject
        return self

    def text(self, content):
        self.text_content = content
        return self

    def html(self, content):
        self.html_content = content
        return self

    def view(self, view, data={}):
        return self.html(
            self.application.make("view").render(view, data).rendered_template
        )

    def priority(self, priority):
        self._priority = str(priority)
        return self

    def high_priority(self):
        self.priority(1)
        return self

    def low_priority(self):
        self.priority(5)
        return self

    def driver(self, driver):
        self._driver = driver
        return self

    def get_response(self):
        self.build()
        if self.get_options().get("html_content"):
            return self.get_options().get("html_content")
        if self.get_options().get("text_content"):
            return self.get_options().get("text_content")

    def get_options(self):
        return {
            "to": self._to,
            "cc": self._cc,
            "bcc": self._bcc,
            "from": self._from,
            "subject": self._subject,
            "text_content": self.text_content,
            "html_content": self.html_content,
            "reply_to": self._reply_to,
            "attachments": self.attachments,
            "priority": self._priority,
            "driver": self._driver,
            "headers": self.headers,
        }

    def build(self, *args, **kwargs):
        return self



"""New Key Command."""
from cryptography.fernet import Fernet

from .Command import Command


class KeyCommand(Command):
    """
    Generate a new key.

    key
        {--s|--store : Stores the key in the .env file}
        {--d|--dont-store : Does not store the key in the .env file}
    """

    def handle(self):
        key = bytes(Fernet.generate_key()).decode("utf-8")

        if self.option("dont-store"):
            return self.info(f"Key: {key}")

        with open(".env", "r") as file:
            data = file.readlines()

        for line_number, line in enumerate(data):
            if line.startswith("APP_KEY="):
                data[line_number] = "APP_KEY={}\n".format(key)
                break

        with open(".env", "w") as file:
            file.writelines(data)

        self.info("Key added to your .env file: {}".format(key))



from tests import TestCase
import os
import time


class TestFileCache(TestCase):
    def setUp(self):
        super().setUp()
        self.application.make("cache")
        self.driver = self.application.make("cache").store()

    def test_can_get_file_driver(self):
        self.driver.put("key", "value")
        self.assertEqual(self.driver.get("key"), "value")
        self.assertTrue(self.driver.has("key"), "value")

    def test_can_add_file_driver(self):
        self.assertEqual(self.driver.add("add_key", "value"), "value")

    def test_can_increment(self):
        self.driver.put("count", "1")
        self.assertEqual(self.driver.get("count"), "1")
        self.driver.increment("count")
        self.assertEqual(self.driver.get("count"), "2")
        self.driver.decrement("count")
        self.assertEqual(self.driver.get("count"), "1")

    def test_will_not_get_expired(self):
        self.driver.put("expire", "1", 1)

        time.sleep(2)
        self.assertEqual(self.driver.get("expire"), None)

    def test_will_get_not_expired(self):
        self.driver.put("expire", "1", 20)
        self.assertEqual(self.driver.get("expire"), "1")

    def test_forget(self):
        self.driver.put("forget", "1")
        self.assertEqual(self.driver.get("forget"), "1")
        self.driver.forget("forget")
        self.assertEqual(self.driver.get("forget"), None)

    def test_remember(self):
        self.driver.remember("remember", lambda cache: (cache.put("remember", "1", 10)))
        self.assertEqual(self.driver.get("remember"), "1")

    def test_remember_datatypes(self):
        self.driver.remember(
            "dic", lambda cache: (cache.put("dic", {"id": 1, "name": "Joe"}, 10))
        )
        self.assertIsInstance(self.driver.get("dic"), dict)
        self.driver.remember("list", lambda cache: (cache.put("list", [1, 2, 3], 10)))
        self.assertIsInstance(self.driver.get("list"), list)

    def test_flush(self):
        self.driver.remember(
            "dic", lambda cache: (cache.put("dic", {"id": 1, "name": "Joe"}, 10))
        )
        self.driver.flush()
        self.assertIsNone(self.driver.get("dic"))



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...foundation import Application
    from ...request import Request
    from ...response import Response


class CookieDriver:
    """Session driver used to store data in HTTP cookies."""

    def __init__(self, application: "Application"):
        self.application = application

    def start(self) -> dict:
        request = self.get_request()
        data = {}
        flashed = {}
        for key, value in request.cookie_jar.to_dict().items():
            if key.startswith("s_"):
                data.update({key.replace("s_", ""): value})
            elif key.startswith("f_"):
                flashed.update({key.replace("f_", ""): value})

        return {"data": data, "flashed": flashed}

    def save(
        self, added=None, deleted=None, flashed=None, deleted_flashed=None
    ) -> None:
        response = self.get_response()
        if added is None:
            added = {}
        if deleted is None:
            deleted = []
        if flashed is None:
            flashed = {}
        if deleted_flashed is None:
            deleted_flashed = []

        for key, value in added.items():
            response.cookie(f"s_{key}", value)

        for key, value in flashed.items():
            response.cookie(f"f_{key}", value)

        for key in deleted:
            response.delete_cookie(f"s_{key}")

        for key in deleted_flashed:
            response.delete_cookie(f"f_{key}")

    def get_response(self) -> "Response":
        return self.application.make("response")

    def get_request(self) -> "Request":
        return self.application.make("request")

    def helper(self) -> "CookieDriver":
        """Use to create builtin helper function."""
        return self



from masonite.controllers import Controller
from masonite.views import View


class __class__(Controller):
    def index(self, view: View):
        return view.render("")

    def show(self, view: View):
        return view.render("")

    def store(self, view: View):
        return view.render("")

    def update(self, view: View):
        return view.render("")

    def destroy(self, view: View):
        return view.render("")



from masoniteorm.connections import ConnectionResolver

DATABASES = {
    "default": "sqlite",
    "sqlite": {
        "driver": "sqlite",
        "database": "database.sqlite3",
    },
}

DB = ConnectionResolver().set_connection_details(DATABASES)



"""Scaffold Auth Command."""
import os
from ...utils.filesystem import render_stub_file, get_module_dir
from ...utils.location import config_path
from ...utils.str import random_string


from ...commands.Command import Command


class APIInstallCommand(Command):
    """
    Adds required files for building API's

    api:install
        {--f|force=? : Force overriding file if already exists}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        stub_path = self.get_stub_config_path()

        content = render_stub_file(stub_path, "api.py")

        path = config_path("api.py")
        if os.path.exists(path) and not self.option("force"):
            self.warning(
                f"{path} already exists! Run the command with -f (force) to override."
            )
            return -1

        with open(path, "w") as f:
            f.write(content)

        if os.path.exists(path) and not self.option("force"):
            self.warning(
                f"{path} already exists! Run the command with -f (force) to override."
            )
            return -1

        with open(path, "w") as f:
            f.write(content)

        secret = random_string(25)

        self.info(f"API Installed: ({config_path('api.py', absolute=False)})")
        self.info(
            f"JWT Secret Key Is: {secret}. You should store this in an environment variable called JWT_SECRET."
        )

    def get_stub_config_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/api.py")



from masonite.providers import Provider


class __class__(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        pass

    def boot(self):
        pass



from tests import TestCase
from src.masonite.mail import Mailable
from src.masonite.mail.Recipient import Recipient


class Welcome(Mailable):
    def build(self):
        return (
            self.to("idmann509@gmail.com")
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
            .html("<h1>Hello from Masonite!</h1>")
            .driver("smtp")
        )


class ViewMailable(Mailable):
    def build(self):
        return (
            self.to("idmann509@gmail.com")
            .subject("Masonite 4")
            .view("mailables.welcome", {})
        )


class TestMailable(TestCase):
    def setUp(self):
        super().setUp()
        self.application.make("mail")

    def test_build_mail(self):
        mailable = Welcome().build().get_options()
        self.assertEqual(mailable.get("to"), "idmann509@gmail.com")
        self.assertEqual(mailable.get("from"), "joe@masoniteproject.com")
        self.assertEqual(mailable.get("subject"), "Masonite 4")
        self.assertEqual(mailable.get("text_content"), "Hello from Masonite!")
        self.assertEqual(mailable.get("html_content"), "<h1>Hello from Masonite!</h1>")
        self.assertEqual(mailable.get("reply_to"), "")
        self.assertEqual(mailable.get("driver"), "smtp")

    def test_build_mailable_view(self):
        mailable = (
            ViewMailable().set_application(self.application).build().get_options()
        )
        self.assertEqual(mailable.get("html_content"), "<h1>Welcome Email</h1>")
        mailable = ViewMailable().set_application(self.application).get_response()
        self.assertEqual(mailable, "<h1>Welcome Email</h1>")

    def test_attach(self):
        self.assertTrue(
            len(
                Welcome()
                .attach("invoice", "tests/integrations/storage/invoice.pdf")
                .build()
                .get_options()
                .get("attachments")
            )
            == 1
        )

    def test_recipient(self):
        to = Recipient("idmann509@gmail.com, joe@masoniteproject.com")
        self.assertEqual(
            to.header(), "<idmann509@gmail.com>, <joe@masoniteproject.com>"
        )
        to = Recipient("Joseph Mancuso <idmann509@gmail.com>, joe@masoniteproject.com")
        self.assertEqual(
            to.header(),
            "Joseph Mancuso <idmann509@gmail.com>, <joe@masoniteproject.com>",
        )

        to = Recipient(
            ["Joseph Mancuso <idmann509@gmail.com>", "joe@masoniteproject.com"]
        )
        self.assertEqual(
            to.header(),
            "Joseph Mancuso <idmann509@gmail.com>, <joe@masoniteproject.com>",
        )



import os
from ..app.User import User

GUARDS = {
    "default": "web",
    "web": {"model": User},
    "password_reset_table": "password_resets",
    "password_reset_expiration": 1440,  # in minutes. 24 hours. None if disabled
}



from tests import TestCase
from src.masonite.response import Response
from src.masonite.routes import Route


class TestResponseRedirect(TestCase):
    def setUp(self):
        super().setUp()
        self.addRoutes(Route.get("/", None).name("home-redirect"))
        self.response = Response(self.application)

    def test_redirect(self):
        self.response.redirect("/")
        self.assertEqual(self.response.get_status(), 302)
        self.assertEqual(self.response.header_bag.get("Location").value, "/")

        self.response.redirect("/", query_params={"key": "value"})
        self.assertEqual(self.response.header_bag.get("Location").value, "/?key=value")

    def test_redirect_to_route_named_route(self):
        self.response.redirect(name="home-redirect")
        self.assertEqual(self.response.get_status(), 302)
        self.assertEqual(self.response.header_bag.get("Location").value, "/")

        self.response.redirect(name="home-redirect", query_params={"key": "value"})
        self.assertEqual(self.response.header_bag.get("Location").value, "/?key=value")

    def test_redirect_to_url(self):
        self.response.redirect(url="/login")
        self.assertEqual(self.response.get_status(), 302)
        self.assertEqual(self.response.header_bag.get("Location").value, "/login")

        self.response.redirect(url="/login", query_params={"key": "value"})
        self.assertEqual(
            self.response.header_bag.get("Location").value, "/login?key=value"
        )



OPTIONS = {
    "editor": "vscode",
    "search_url": "https://www.google.com/search?q=",
    "links": {
        "doc": "https://docs.masoniteproject.com",
        "repo": "https://github.com/MasoniteFramework/masonite",
    },
    "stack": {"offset": 10, "shorten": True},
    "hide_sensitive_data": True,
}

HANDLERS = {
    "dumps": True,
    "solutions": {"stackoverflow": False, "possible_solutions": True},
    "recommendations": {
        "packages_updates": {
            "list": ["exceptionite", "masonite", "masonite-orm", "pytest"]
        }
    },
}



"""Password Helper Module."""
import uuid
from ...facades import Hash


class Authenticates:

    __auth__ = "email"
    __password__ = "password"

    def attempt(self, username, password):
        """Attempts to login using a username and password"""
        record = self.where(self.get_username_column(), username).first()
        if not record:
            return False

        record_password = getattr(record, self.get_password_column())
        if not isinstance(record_password, bytes):
            record_password = bytes(record_password or "", "utf-8")
        try:
            if Hash.check(password, record_password):
                record.set_remember_token().save()
                return record
        except ValueError:
            pass

        return False

    def get_auth_column(self, username):
        record = self.where(self.get_username_column(), username).first()
        if not record:
            return False

        return getattr(record, self.get_username_column())

    def register(self, dictionary):
        dictionary.update(
            {self.get_password_column(): Hash.make(dictionary.get("password", ""))}
        )
        return self.create(dictionary)

    def get_id(self):
        return self.get_primary_key_value()

    def attempt_by_id(self, user_id):
        """Attempts to login using a username and password"""
        record = self.find(user_id)
        if not record:
            return False

        record.set_remember_token().save()
        return record

    def get_remember_token(self):
        """Attempts to login using a username and password"""
        return self.remember_token

    def set_remember_token(self, token=None):
        """Attempts to login using a username and password"""
        self.remember_token = str(token) if token else str(uuid.uuid4())
        return self

    def reset_password(self, username, password):
        """Attempts to login using a username and password"""
        self.where(self.get_username_column(), username).update(
            {self.get_password_column(): Hash.make(password)}
        )
        return self

    def get_password_column(self):
        """Attempts to login using a username and password"""
        return self.__password__

    def get_username_column(self):
        """Attempts to login using a username and password"""
        return self.__auth__



class StorageCapsule:
    def __init__(self):
        self.storage_templates = {}

    def add_storage_assets(self, templates):
        self.storage_templates.update(templates)
        return self

    def get_storage_assets(self):
        return self.storage_templates



from src.masonite.environment import env


FROM_EMAIL = env("MAIL_FROM", "no-reply@masonite.com")

DRIVERS = {
    "default": env("MAIL_DRIVER", "terminal"),
    "smtp": {
        "host": env("MAIL_HOST"),
        "port": env("MAIL_PORT", "587"),
        "username": env("MAIL_USERNAME"),
        "password": env("MAIL_PASSWORD"),
        "from": FROM_EMAIL,
    },
    "mailgun": {
        "domain": env("MAILGUN_DOMAIN"),
        "secret": env("MAILGUN_SECRET"),
        "from": FROM_EMAIL,
    },
    "terminal": {
        "from": FROM_EMAIL,
    },
}



from ...providers import Provider
from ..Broadcast import Broadcast
from ..drivers import PusherDriver
from ...configuration import config


class BroadcastProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        broadcast = Broadcast(self.application).set_configuration(
            config("broadcast.broadcasts")
        )
        broadcast.add_driver("pusher", PusherDriver(self.application))

        self.application.bind("broadcast", broadcast)

    def boot(self):
        pass



import unittest
from src.masonite.validation import MessageBag


class TestMessageBag(unittest.TestCase):
    def setUp(self):
        self.bag = MessageBag()

    def test_message_bag_can_add_errors_and_messages(self):
        self.bag.add("email", "Your email is invalid")
        self.assertEqual(self.bag.items, {"email": ["Your email is invalid"]})

    def test_message_bag_can_add_several_errors_and_messages(self):
        self.bag.add("email", "Your email is invalid")
        self.bag.add("email", "Your email is invalid")

    def test_message_bag_can_get_all_errors_and_messages(self):
        self.bag.reset()
        self.bag.add("email", "Your email is invalid")
        self.assertEqual(self.bag.all(), {"email": ["Your email is invalid"]})

    def test_message_bag_has_any_errors(self):
        self.bag.reset()
        self.assertFalse(self.bag.any())
        self.bag.add("email", "Your email is invalid")
        self.assertTrue(self.bag.any())

    def test_message_bag_has_any_errors(self):
        self.bag.reset()
        self.assertTrue(self.bag.empty())
        self.bag.add("email", "Your email is invalid")
        self.assertFalse(self.bag.empty())

    def test_message_bag_can_get_first_error(self):
        self.bag.reset()
        self.bag.add("email", "Your email is invalid")
        self.bag.add("email", "Your email is too short")
        self.bag.add("username", "Your username is invalid")
        self.assertEqual(self.bag.first("email"), "Your email is invalid")
        self.assertEqual(self.bag.first("username"), "Your username is invalid")

    def test_amount_of_messages(self):
        self.bag.reset()
        self.bag.add("email", "Your email is invalid")
        self.bag.add("email", "Your email too short")
        self.assertEqual(self.bag.amount("email"), 2)

    def test_has_message(self):
        self.bag.reset()
        self.assertFalse(self.bag.has("email"))
        self.assertFalse(self.bag.has("username"))
        self.bag.add("email", "Your email is invalid")
        self.bag.add("email", "Your email too short")
        self.assertTrue(self.bag.has("email"))
        self.assertFalse(self.bag.has("username"))

    def test_get_messages_with_same_keys(self):
        self.bag.reset()
        self.bag.add("email", "Your email is invalid")
        self.bag.add("email", "Your email too short")
        self.assertIn(
            "Your email is invalid",
            self.bag.get("email"),
        )
        self.assertIn(
            "Your email too short",
            self.bag.get("email"),
        )

    def test_get_errors(self):
        self.bag.reset()
        self.bag.add("email", "Your email is invalid")
        self.bag.add("email", "Your email too short")
        self.assertEqual(self.bag.errors(), ["email"])

    def test_get_messages(self):
        self.bag.reset()
        self.bag.add("email", "Your email is invalid")
        self.bag.add("username", "Your username too short")
        self.assertIn(
            "Your email is invalid",
            self.bag.messages(),
        )

        self.assertIn(
            "Your username too short",
            self.bag.messages(),
        )

    def test_can_convert_to_json(self):
        self.bag.reset()
        self.bag.add("email", "Your email is invalid")
        self.assertEqual(self.bag.json(), '{"email": ["Your email is invalid"]}')

    def test_can_merge(self):
        self.bag.reset()
        self.bag.add("email", "Your email is invalid")
        self.bag.merge({"username": ["username is too short"]})
        self.assertEqual(self.bag.count(), 2)

    def test_can_work_with_if_statements_and_full(self):
        self.bag.reset()
        self.bag.add("email", "Your email is invalid")

        if self.bag:
            pass
        else:
            raise AssertionError("Should assert true as a boolean")

    def test_can_work_with_if_statements_and_false(self):
        self.bag.reset()

        if self.bag:
            raise AssertionError("Should not raise when not full")
        else:
            pass



"""This module is overriding ShellCommand to rename to db:shell to make more sense
than shell inside a Masonite project."""
from masoniteorm.commands import ShellCommand


class DbShellCommand(ShellCommand):
    """
    Connect to your database interactive terminal.

    db:shell
        {--c|connection=default : The connection you want to use to connect to interactive terminal}
        {--s|show=? : Display the command which will be called to connect}
    """



from typing import Any

class Loader:
    def get_modules(files_or_directories: list, raise_exception: bool = False) -> dict:
        """Get a list of Python modules found (recursively) in the given list of files or directories.
        If raise_exception is enabled it will raise an exception in case of error during loading a module."""
    def find(
        class_instance: Any,
        paths: list,
        class_name: str,
        raise_exception: bool = False,
    ) -> "None|Any": ...
    def find_all(
        class_instance: Any, paths: list, raise_exception: bool = False
    ) -> dict: ...
    def get_object(
        path_or_module: "str|Any", object_name: str, raise_exception: bool = False
    ) -> Any:
        """Load the given object from a Python module located at path and returns a default value if
        not found. If no object name is provided, returns the loaded module."""
        ...
    def get_objects(
        path_or_module: "str|Any",
        filter_method: callable = None,
        raise_exception: bool = False,
    ) -> dict:
        """Returns a dictionary of objects from the given path (file or dotted). The dictionary can
        be filtered if a given callable is given."""
        ...
    def get_parameters(module_or_path: "str|Any") -> dict: ...



from ..facades import Api


class AuthenticatesTokens:

    __TOKEN_COLUMN__ = "api_token"

    def generate_jwt(self):
        token = Api.generate_token()

        setattr(self, self.__TOKEN_COLUMN__, token)
        self.save()
        return token

    def attempt_by_token(self, token):
        return self.where(self.__TOKEN_COLUMN__, token).first()



from tests import TestCase

from src.masonite.utils.filesystem import get_extension


class TestFileUtils(TestCase):
    def test_get_extension(self):
        self.assertEqual(get_extension("log.txt"), ".txt")
        self.assertEqual(get_extension("archive.tar.gz"), ".tar.gz")
        self.assertEqual(get_extension("path/to/log.txt"), ".txt")

        self.assertEqual(get_extension("log.txt", without_dot=True), "txt")
        self.assertEqual(get_extension("archive.tar.gz", without_dot=True), "tar.gz")



from ..app.User import User

DRIVERS = {
    "jwt": {
        "algorithm": "HS512",
        "secret": "secret",
        "model": User,
        "expires": None,
        "authenticates": False,
        "version": None,
    }
}



from tests import TestCase
from src.masonite.facades import Config


class TestApplication(TestCase):
    def test_is_running_tests(self):
        self.assertTrue(self.application.is_running_tests())

    def test_is_debug(self):
        original_value = Config.get("application.debug")

        Config.set("application.debug", True)
        self.assertTrue(self.application.is_debug())
        Config.set("application.debug", False)
        self.assertFalse(self.application.is_debug())

        Config.set("application.debug", original_value)

    def test_is_production(self):
        with self.env("production"):
            self.assertTrue(self.application.is_production())



from tests import TestCase

from src.masonite.utils.http import generate_wsgi
from src.masonite.request import Request


class TestRequest(TestCase):
    def setUp(self):
        super().setUp()
        self.request = Request(generate_wsgi(path="/test"))

    def test_request_no_input_returns_empty_string(self):
        self.assertEqual(self.request.input("notavailable"), "")

    def test_request_can_get_string_value(self):
        self.request.input_bag.query_string = {"test": "value"}
        self.assertEqual(self.request.input("test"), "value")

    def test_request_can_get_list_value(self):
        self.request.input_bag.query_string = {"test": ["foo", "bar"]}
        self.assertEqual(self.request.input("test"), ["foo", "bar"])

    def test_request_can_get_float_value(self):
        self.request.input_bag.query_string = {"test": 3.1415926}
        self.assertEqual(self.request.input("test"), 3.1415926)

    def test_request_can_get_nested_value_from_query_string(self):
        self.request.input_bag.query_string = {"key": "val", "a": {"b": {"c": 1}}}
        self.assertEqual(self.request.input("a.b.c"), 1)

    def test_request_can_get_nested_value_from_post_data(self):
        self.request.input_bag.post_data = {"key": "val", "a": {"b": {"c": 1}}}
        self.assertEqual(self.request.input("a.b.c"), 1)



from src.masonite.providers import (
    RouteProvider,
    FrameworkProvider,
    ViewProvider,
    WhitenoiseProvider,
    ExceptionProvider,
    MailProvider,
    SessionProvider,
    QueueProvider,
    CacheProvider,
    EventProvider,
    StorageProvider,
    HelpersProvider,
    BroadcastProvider,
    AuthenticationProvider,
    AuthorizationProvider,
    HashServiceProvider,
    ORMProvider,
)


from src.masonite.scheduling.providers import ScheduleProvider
from src.masonite.notification.providers import NotificationProvider
from src.masonite.validation.providers.ValidationProvider import ValidationProvider
from src.masonite.api.providers import ApiProvider
from ..test_package import MyTestPackageProvider

from tests.integrations.providers import AppProvider

PROVIDERS = [
    FrameworkProvider,
    HelpersProvider,
    RouteProvider,
    ViewProvider,
    WhitenoiseProvider,
    ExceptionProvider,
    MailProvider,
    NotificationProvider,
    SessionProvider,
    CacheProvider,
    QueueProvider,
    ScheduleProvider,
    EventProvider,
    StorageProvider,
    BroadcastProvider,
    HashServiceProvider,
    AuthenticationProvider,
    AuthorizationProvider,
    ValidationProvider,
    MyTestPackageProvider,
    AppProvider,
    ORMProvider,
    ApiProvider,
]



from tests import TestCase


class TestCookieSession(TestCase):
    def test_can_start_session(self):
        request = self.make_request()
        session = self.application.make("session")
        request.cookie("s_hello", "test")
        session.start()
        self.assertEqual(session.get("hello"), "test")

    def test_can_get_session_dict(self):
        request = self.make_request()
        session = self.application.make("session")
        request.cookie("s_hello", '{"hello": "test"}')
        session.start()
        self.assertEqual(type(session.get("hello")), dict)

    def test_can_set_and_get_session_dict(self):
        request = self.make_request()
        session = self.application.make("session")
        session.start()
        session.set("key1", {"hello": "test"})
        self.assertEqual(type(session.get("key1")), dict)
        self.assertEqual(session.get("key1")["hello"], "test")

    def test_can_set_and_get_session_dict(self):
        request = self.make_request()
        session = self.application.make("session")
        session.start()
        session.flash("key1", {"hello": "test"})
        self.assertEqual(session.get("key1")["hello"], "test")

    def test_can_set_and_get_session(self):
        self.make_request()
        session = self.application.make("session")
        session.start()
        session.set("key1", "test1")
        self.assertEqual(session.get("key1"), "test1")

    def test_can_increment_and_decrement_session(self):
        self.make_request()
        session = self.application.make("session")
        session.start()
        session.set("key1", "1")
        session.set("key5", "5")
        session.increment("key1")
        session.decrement("key5")
        self.assertEqual(session.get("key1"), "2")
        self.assertEqual(session.get("key5"), "4")

    def test_can_save_session(self):
        self.make_request()
        response = self.make_response()
        session = self.application.make("session")
        session.start()
        session.set("key1", "test1")

        self.assertEqual(response.cookie("s_key1"), "test1")

    def test_can_delete_session(self):
        request = self.make_request()
        response = self.make_response()
        session = self.application.make("session")
        request.cookie("s_key", "test")
        session.start()

        self.assertEqual(session.get("key"), "test")

        session.delete("key")
        self.assertEqual(session.get("key"), None)

        self.assertEqual(response.cookie("s_key"), None)
        self.assertTrue("s_key" in response.cookie_jar.deleted_cookies)

    def test_can_pull_session(self):
        request = self.make_request()
        response = self.make_response()
        session = self.application.make("session")
        request.cookie("s_key", "test")
        session.start()

        self.assertEqual(session.get("key"), "test")

        key = session.pull("key")
        self.assertEqual(key, "test")
        self.assertEqual(session.get("key"), None)
        self.assertEqual(response.cookie("s_key"), None)
        self.assertTrue("s_key" in response.cookie_jar.deleted_cookies)

    def test_can_flush_session(self):
        request = self.make_request()
        response = self.make_response()
        session = self.application.make("session")
        request.cookie("s_key", "test")
        session.start()

        self.assertEqual(session.get("key"), "test")

        session.flush()
        self.assertEqual(session.get("key"), None)
        self.assertEqual(response.cookie("s_key"), None)
        self.assertTrue("s_key" in response.cookie_jar.deleted_cookies)

    def test_can_flash(self):
        request = self.make_request()
        response = self.make_response()
        session = self.application.make("session")
        session.start()
        session.flash("key", "test")
        self.assertEqual(session.get("key"), "test")
        self.assertEqual(session.get("key"), None)
        self.assertEqual(response.cookie("f_key"), None)

    def test_flash_two_keys_does_not_duplicate_data(self):
        request = self.make_request()
        response = self.make_response()
        session = self.application.make("session")
        session.start()
        session.flash("key", "test")
        session.flash("key2", "test2")

        self.assertTrue(session.has("key"))
        self.assertTrue(session.has("key2"))
        self.assertTrue(response.cookie_jar.exists("f_key"))
        self.assertTrue(response.cookie_jar.exists("f_key2"))

        self.assertEqual(session.get("key"), "test")

        self.assertFalse(session.has("key"))
        self.assertTrue(session.has("key2"))
        self.assertFalse(response.cookie_jar.exists("f_key"))
        self.assertTrue(response.cookie_jar.exists("f_key2"))



from ...facades import Gate


class Authorizes:
    def can(self, permission, *args):
        return Gate.for_user(self).allows(permission, *args)

    def cannot(self, permission, *args):
        return Gate.for_user(self).denies(permission, *args)

    def can_any(self, permissions, *args):
        return Gate.for_user(self).any(permissions, *args)



import builtins
from markupsafe import Markup

from ..providers import Provider
from ..configuration import config
from ..helpers import UrlsHelper, MixHelper, optional


class HelpersProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        builtins.resolve = self.application.resolve
        builtins.container = lambda: self.application
        self.application.bind("url", UrlsHelper(self.application))

    def boot(self):
        request = self.application.make("request")
        urls_helper = self.application.make("url")

        self.application.make("view").share(
            {
                "request": lambda: request,
                "session": lambda: request.app.make("session"),
                "auth": request.user,
                "cookie": request.cookie,
                "back": lambda url=request.get_path_with_query(): (
                    Markup(f"<input type='hidden' name='__back' value='{url}' />")
                ),
                "method": lambda method: (
                    Markup(f"<input type='hidden' name='__method' value='{method}' />")
                ),
                "asset": urls_helper.asset,
                "url": urls_helper.url,
                "mix": MixHelper(self.application).url,
                "route": urls_helper.route,
                "config": config,
                "can": self.application.make("gate").allows,
                "cannot": self.application.make("gate").denies,
                "optional": optional,
                "exists": self.application.make("view").exists,
            }
        )



from src.masonite.authorization import Policy


class PostPolicy(Policy):
    def create(self, user):
        return user.email == "idmann509@gmail.com"

    def view_any(self, user=None):
        """Here we allow guests (non-authenticated) users by declaring the user optional else
        only authenticated users would have been allowed."""
        return True

    def view(self, user, instance):
        return False

    def update(self, user, instance):
        return user.id == instance.user_id

    def delete(self, user, instance):
        if user.id != instance.user_id:
            return self.deny("You do not own this post")
        else:
            return self.allow()

    def force_delete(self, user, instance):
        return False

    def restore(self, user, instance):
        return False



from src.masonite.controllers import Controller
from src.masonite.views import View
from tests.integrations.app.User import User
from src.masonite.request import Request
from src.masonite.response import Response
from src.masonite.authentication import Auth


class PasswordResetController(Controller):
    def show(self, view: View):  # Show password_reset page
        return view.render("auth.password_reset")

    def store(
        self, auth: Auth, request: Request, response: Response
    ):  # store password_reset record
        auth.password_reset(request.input("email"))
        return "event fired"

    def change_password(self, view: View):  # store password_reset record
        return view.render("auth.change_password")

    def store_changed_password(
        self, auth: Auth, request: Request, response: Response
    ):  # store password_reset record
        auth.reset_password(request.input("password"), request.input("token"))

        # Need to validate??
        # Redirect back?
        return response.back()



from tests import TestCase
from masoniteorm.models import Model

from tests.integrations.policies.PostPolicy import PostPolicy
from src.masonite.exceptions.exceptions import PolicyDoesNotExist


class User(Model):
    """User Model"""

    __fillable__ = ["name", "email", "password"]


class Post(Model):
    __fillable__ = ["user_id", "name"]


class TestPolicies(TestCase):
    def setUp(self):
        super().setUp()
        self.gate = self.application.make("gate")
        self.make_request()

    def tearDown(self):
        super().tearDown()
        self.gate.policies = {}
        self.gate.permissions = {}

    def test_can_register_policies(self):
        self.gate.register_policies([(Post, PostPolicy)])
        self.assertEqual(self.gate.policies[Post], PostPolicy)

    def test_using_policies_with_argument(self):
        self.gate.register_policies([(Post, PostPolicy)])
        post = Post()
        post.user_id = 1
        # authenticates user 1
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        self.assertTrue(self.gate.allows("update", post))

    def test_using_policies_without_argument(self):
        self.gate.register_policies([(Post, PostPolicy)])
        # authenticates user 1
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")

        self.assertTrue(self.gate.allows("create", Post))

    def test_using_policy_returning_response(self):
        self.gate.register_policies([(Post, PostPolicy)])
        # authenticates user 1
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        post = Post()
        post.user_id = 3
        response = self.gate.inspect("delete", post)
        self.assertFalse(response.allowed())
        self.assertEqual(response.message(), "You do not own this post")

    def test_that_use_defined_gate_if_no_policy_match(self):
        self.gate.define("update", lambda user, post: user.id == post.user_id)
        post = Post()
        post.user_id = 1
        # authenticates user 1
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        # here no policy has been defined, the gate defined above will be used
        self.assertTrue(self.gate.allows("update", post))

    def test_that_policy_can_allow_guest_users(self):
        self.gate.register_policies([(Post, PostPolicy)])
        self.assertTrue(self.gate.allows("view_any", Post))

    def test_any_on_policy(self):
        self.gate.register_policies([(Post, PostPolicy)])
        post = Post()
        post.user_id = 1
        # authenticates user 1
        self.application.make("auth").attempt("idmann509@gmail.com", "secret")
        self.assertTrue(self.gate.any(["update", "delete"], post))

    def test_unknown_policy_method_raises_exception(self):
        self.gate.register_policies([(Post, PostPolicy)])
        with self.assertRaises(PolicyDoesNotExist):
            self.gate.allows("can-fly", Post)



import pytest
import responses
from tests import TestCase
from src.masonite.notification import Notification, Notifiable, SlackMessage
from src.masonite.exceptions import NotificationException

from masoniteorm.models import Model

# fake webhook for tests
webhook_url = "https://hooks.slack.com/services/X/Y/Z"
webhook_url_2 = "https://hooks.slack.com/services/A/B/C"


def route_for_slack(self):
    return "#bot"


class User(Model, Notifiable):
    """User Model"""

    __fillable__ = ["name", "email", "password", "phone"]

    def route_notification_for_slack(self):
        return route_for_slack(self)


class WelcomeUserNotification(Notification):
    def to_slack(self, notifiable):
        return SlackMessage().text(f"Welcome {notifiable.name}!").from_("test-bot")

    def via(self, notifiable):
        return ["slack"]


class WelcomeNotification(Notification):
    def to_slack(self, notifiable):
        return SlackMessage().text("Welcome !").from_("test-bot")

    def via(self, notifiable):
        return ["slack"]


class OtherNotification(Notification):
    def to_slack(self, notifiable):
        return (
            SlackMessage().to(["#general", "#news"]).text("Welcome !").from_("test-bot")
        )

    def via(self, notifiable):
        return ["slack"]


class TestSlackWebhookDriver(TestCase):
    def setUp(self):
        super().setUp()
        self.notification = self.application.make("notification")

    @responses.activate
    def test_sending_to_anonymous(self):
        responses.add(responses.POST, webhook_url, body=b"ok")
        self.notification.route("slack", webhook_url).notify(WelcomeNotification())
        self.assertTrue(responses.assert_call_count(webhook_url, 1))

    @responses.activate
    def test_sending_to_notifiable(self):
        responses.add(responses.POST, webhook_url, body=b"ok")
        User.route_notification_for_slack = lambda notifiable: webhook_url
        user = User.find(1)
        user.notify(WelcomeNotification())
        self.assertTrue(responses.assert_call_count(webhook_url, 1))
        User.route_notification_for_slack = route_for_slack

    @responses.activate
    def test_sending_to_multiple_webhooks(self):
        responses.add(responses.POST, webhook_url, body=b"ok")
        responses.add(responses.POST, webhook_url_2, body=b"ok")
        User.route_notification_for_slack = lambda notifiable: [
            webhook_url,
            webhook_url_2,
        ]
        user = User.find(1)
        user.notify(WelcomeNotification())
        self.assertTrue(responses.assert_call_count(webhook_url, 1))
        self.assertTrue(responses.assert_call_count(webhook_url_2, 1))
        User.route_notification_for_slack = route_for_slack


class TestSlackAPIDriver(TestCase):
    url = "https://slack.com/api/chat.postMessage"
    channel_url = "https://slack.com/api/conversations.list"

    def setUp(self):
        super().setUp()
        self.notification = self.application.make("notification")

    def test_sending_without_credentials(self):
        with self.assertRaises(NotificationException) as e:
            self.notification.route("slack", "123456").notify(WelcomeNotification())
        self.assertIn("not_authed", str(e.exception))

    @responses.activate
    def test_sending_to_anonymous(self):
        responses.add(
            responses.POST,
            self.url,
            body=b'{"ok": "True"}',
        )
        responses.add(
            responses.POST,
            self.channel_url,
            body=b'{"channels": [{"name": "bot", "id": "123"}]}',
        )
        self.notification.route("slack", "#bot").notify(WelcomeNotification())
        # to convert #bot to Channel ID
        self.assertTrue(responses.assert_call_count(self.channel_url, 1))
        self.assertTrue(responses.assert_call_count(self.url, 1))

    @responses.activate
    def test_sending_to_notifiable(self):
        user = User.find(1)
        responses.add(
            responses.POST,
            self.url,
            body=b'{"ok": "True"}',
        )
        responses.add(
            responses.POST,
            self.channel_url,
            body=b'{"channels": [{"name": "bot", "id": "123"}]}',
        )
        user.notify(WelcomeUserNotification())
        self.assertTrue(responses.assert_call_count(self.url, 1))

    @responses.activate
    @pytest.mark.skip(
        reason="Failing because user defined routing takes precedence. What should be the behaviour ?"
    )
    def test_sending_to_multiple_channels(self):
        user = User.find(1)
        responses.add(
            responses.POST,
            self.url,
            body=b'{"ok": "True"}',
        )
        responses.add(
            responses.POST,
            self.channel_url,
            body=b'{"channels": [{"name": "bot", "id": "123"}, {"name": "general", "id": "456"}]}',
        )
        user.notify(OtherNotification())
        self.assertTrue(responses.assert_call_count(self.channel_url, 2))
        self.assertTrue(responses.assert_call_count(self.url, 2))

    @responses.activate
    def test_convert_channel(self):
        channel_id = self.notification.get_driver("slack").convert_channel(
            "123456", "token"
        )
        self.assertEqual(channel_id, "123456")

        responses.add(
            responses.POST,
            self.channel_url,
            body=b'{"channels": [{"name": "general", "id": "654321"}]}',
        )
        channel_id = self.notification.get_driver("slack").convert_channel(
            "#general", "token"
        )
        self.assertEqual(channel_id, "654321")



from tests import TestCase
from src.masonite.mail import Mail, MockMail


class CustomMockMail:
    def __init__(self, application):
        self.application = application


class TestMocking(TestCase):
    def test_fake_service(self):
        mocked_mail = self.fake("mail")

        assert isinstance(self.application.make("mail"), MockMail)
        assert isinstance(mocked_mail, MockMail)

        self.restore("mail")

        assert isinstance(self.application.make("mail"), Mail)



"""A Validation Service Provider."""
from ...providers import Provider
from .. import Validator, ValidationFactory, MessageBag
from ..commands.MakeRuleEnclosureCommand import MakeRuleEnclosureCommand
from ..commands.MakeRuleCommand import MakeRuleCommand


class ValidationProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        validator = Validator()
        self.application.bind("validator", validator)
        self.application.make("commands").add(
            MakeRuleEnclosureCommand(self.application),
            MakeRuleCommand(self.application),
        )

        MessageBag.get_errors = self._get_errors
        self.application.make("view").share({"bag": MessageBag.view_helper})
        validator.extend(ValidationFactory().registry)

    def boot(self):
        pass

    def _get_errors(self):
        request = self.application.make("request")
        messages = []
        for error, message in (
            request.session.get_flashed_messages().get("errors", {}).items()
        ):
            messages += message

        return messages



from cleo import Command
from cleo import Application as CommandApplication
from src.masonite.commands import CommandCapsule
from tests import TestCase


class FakeTestCommand(Command):
    """
    Command for testing command assertions only.
    fake_test_command
        {--f|--fail : Make command fail}
    """

    def handle(self):
        self.info("Command success !")


class TestCommandsAssertions(TestCase):
    def setUp(self):
        super().setUp()
        self.original_commands = self.application.make("commands")
        command_app = CommandCapsule(CommandApplication("Masonite Version:", "tests"))
        self.application.bind("commands", command_app)
        self.application.make("commands").add(FakeTestCommand())

    def tearDown(self):
        super().tearDown()
        self.application.bind("commands", self.original_commands)

    def test_running_command_during_tests(self):
        self.craft("fake_test_command")

    def test_assert_output(self):
        self.craft("fake_test_command").assertOutputContains("Command")
        self.craft("fake_test_command").assertExactOutput("Command success !\n")

    def test_assert_output_missing(self):
        self.craft("fake_test_command").assertOutputMissing(
            "This is not in the command"
        )

    def test_assert_errors(self):
        with self.assertRaises(AssertionError):
            self.craft("fake_test_command").assertHasErrors()

    def test_assert_success(self):
        self.craft("fake_test_command").assertSuccess()



""" __class__ Validation Enclosure """

from masonite.validation import RuleEnclosure


class __class__(RuleEnclosure):
    """__class__ Validation Enclosure Class."""

    def rules(self):
        """Used to return a list of rules in order to make validation
        more reusable.

        Returns:
            list -- List of rules
        """
        return [
            # Rules go here
        ]
